MODULE ParnacIterStepMain

    ! =============================================================================
    !
    !   Programmer: Koos Huijssen / Jasper de Koning
    !
    !   Language: Fortran 90
    !
    !   Version Date    Comment
    !   ------- -----   -------
    !   1.0     090505  Original code (KH)
    !
    !   Copyright (C)   Laboratory of Electromagnetic Research
    !                   Delft University of Technology, Delft, The Netherlands
    !
    !   Permission to copy or distribute this software or documentation
    !   in hard copy or soft copy granted only by written license
    !   obtained from the Laboratory of Electromagnetic Research.
    !   All rights reserved. No parts of this publication may be
    !   reproduced, stored in a retrieval system (e.g. in memory, disk,
    !   or core) or be transmitted by any means, electronic, mechanical,
    !   photocopy, recording, or otherwise, without written permission
    !   from the publisher.
    !
    ! *****************************************************************************
    !
    !   DESCRIPTION
    !
    !   The module ParnacIterStepMain contains subroutines used in the main
    !   program. The larger functions in the main program are implemented here.
    !
    ! *****************************************************************************
    !
    !   MODULES USED
    !
    USE Types
    USE Constants
    USE ParnacGeneral
    USE ParnacDataDef
    USE ParnacParamDef
    USE ParnacIdentifierDef
    !        USE ParnacMPIDef

    USE ParnacDataSourceInit, ONLY: &
        InitSource, InitSource_Planewave, SourceDimensions
    USE ParnacDataSpaceInit, ONLY: &
        InitSpace, InitGrid, DestructSpace, SpaceInfo, &
        GridDistr0CreateEmpty, GridDistr0Deallocate, &
        GridDistr1CreateEmpty, &
        GridDistr2CreateEmpty, &
        iBeamOffsetT, iBeamOffsetX, iBeamOffsetY
    USE ParnacDataMap, ONLY: &
        MapVt, MapVtInv
    USE ParnacContrastFunctions
    USE ParnacConvolution, ONLY: &
        ConvolutionGreen, ConvolutionGreenConj
    USE ParnacTransformFilter, ONLY: &
        TransformT, TransformTInv, TransformT_sml, TransformTInv_sml, &
        TransformXYZ, TransformXYZInv, TransformXYZ_sml, TransformXYZInv_sml, &
        FFilterSpaceSpatial3D, dTaperingWindow
    USE ParnacDataRedist, ONLY: &
        ReorderDistr0ToDistr1, ReorderDistr1ToDistr0, &
        ReorderDistr1ToDistr2, ReorderDistr2ToDistr1
    USE ParnacDerivative, ONLY: &
        BeamDerivativeZ
    USE ParnacOutput, ONLY: &
        ExportSlice, ExportArray
    USE ParnacPointSourceCloud
    ! *****************************************************************************
    !
    !   GLOBAL DECLARATIONS
    !
    !   none
    !
    IMPLICIT NONE

    ! *****************************************************************************
    !
    !   CONTAINED SUBROUTINES AND FUNCTIONS
    !
    !   PrimarySourcetoField    sub   Computes the primary field on the given space
    !                                 that is generated by the primary source
    !
    !   PlanewaveField          sub   Computes a plane wave running through the
    !                                 space in the positive z-direction
    !   FieldtoContrastSource   sub   Compute the contrast source for the given
    !                                 field space
    !   ContrastSourcetoField   sub   Compute the contrast field in a given space
    !                                 from a given contrast source
    !   CFPlaneSourcetoField    sub   Compute the field excited by a plane on
    !                                 which a pressure is defined - more or less
    !                                 Huygens principle (however, surface is not
    !                                 closed, just a part of the XY-plane)
    !   StoreField              sub   Store the field in a given space to a
    !                                 temporary file
    !   LoadField               sub   Load a field from a temporary file to a space
    !   AddStoredToField        sub   Add the field in a space to the field in a
    !                                 temporary file. Put the result in the space
    !   StorePlane              sub   Store a plane from a space to a temporary
    !                                 file
    !   LoadPlane               sub   Load a plane from a temporary file to a space
    !   ZeroFirstPlane          sub   Set the first plane in a space to zero
    !   StoreSlices             sub   Export the slices from a space to file
    !   UpdateRRMSerror         sub   Update the RRMS error for the current
    !                                 iteration and beam
    !   ExportRRMSerror         sub   Export the RRMS error to file
    !   CopyDataD1              sub   Copy the data from one space to a space with
    !                                 a smaller time dimension (Dist. 0)
    !   CopyDataD2              sub   Copy the data from one space to a space with
    !                                 smaller spatial dimensions (Dist. 2)
    !   test_isnan              sub   Test whether there is a NaN in a space
    !
    ! =============================================================================
    !
CONTAINS

    SUBROUTINE PrimarySourcetoField(cSpace)

        ! =============================================================================
        !
        !   Programmer: Koos Huijssen / Jasper de Koning
        !
        !   Language: Fortran 90
        !
        !   Version Date    Comment
        !   ------- -----   -------
        !   1.0     090505  Original code (KH)
        !
        ! *****************************************************************************
        !
        !   DESCRIPTION
        !
        !   The subroutine PrimarySourcetoField computes the primary field on the given
        !   space that is generated by the primary source.
        !
        ! *****************************************************************************
        !
        !   INPUT/OUTPUT PARAMETERS
        !
        !   cSpace   io   type(space)  Space structure for which the primary field
        !                              solution is to be obtained
        !
        type(Space), intent(inout)::                cSpace; 
        ! *****************************************************************************
        !
        !   LOCAL PARAMETERS
        !
        !   cSource         type(Space)   Space containing the primary source
        !   iMarginStartZ   i8b           Start margin in Z for d/dz operation
        !   iMarginStartT   i8b           Start margin in T for d/dz operation
        !   iSourceDimT     i8b           Dimension of the source in T
        !   iSourceDimX     i8b           Dimension of the source in X
        !   iSourceDimY     i8b           Dimension of the source in Y
        !   cSpaceMargin    type(Space)   Space for the field with margins for d/dz
        !   cSpaceXYZMargin type(Space)   Space for the field with margins in XYZ only
        !   aiMargin        i8b           Margins in all four dimensions T/X/Y/Z
        !   dTempPhaseDelay dp            Maximum phase delay in the source
        !   acTemp          char          Temporary char array for output log messages
        !
        type(Space) ::                                cSource
        integer(i8b)::              iMarginStartZ, iMarginStartT, i
        integer(i8b) ::                                iSourceDimT, iSourceDimX, iSourceDimY
        type(Space)::               cSpaceMargin, cSpaceXYZMargin
        integer(i8b) ::             aiMargin(4), iLen
        real(dp) ::                                        dTempPhaseDelay
        character(len=1024)::       acTemp

        !Filtering and windowing parameters
        integer(i8b)                ::        iDimW, iStart, iIndex, iOmega
        real(dp)                         ::        dDOmega, dFFTFactor
        real(dp), allocatable         ::        dTaperSupportWindow(:), dTaperMaxFreqWindow(:), dMultFactor(:)
        real(dp)                         ::        dLeftBand, dRightBand, dCorrection

        ! *****************************************************************************
        !
        !   I/O
        !
        !   log file entries
        !
        ! *****************************************************************************
        !
        !   SUBROUTINES/FUNCTIONS CALLED
        !
        !   SwStartAndCount
        !   PrintToLog
        !   iBeamOffsetT
        !   iBeamOffsetX
        !   iBeamOffsetY
        !   SourceDimensions
        !   InitSpace
        !   InitGrid
        !   SpaceInfo
        !   InitSource
        !   ExportSlice
        !   ExportArray
        !   ConvolutionGreen
        !   DestructSpace
        !   BeamDerivativeZ
        !   GridDistr0CreateEmpty
        !   CopyDataD1
        !   CopyDataD2
        !   ReorderDistr0ToDistr1
        !   ReorderDistr1ToDistr2
        !   ReorderDistr2ToDistr1
        !   ReorderDistr1ToDistr0
        !   GridDistr2CreateEmpty
        !   SWStop
        !
        ! *****************************************************************************
        !
        !   PSEUDOCODE
        !
        !   If the source is an F-source and the aperture is not a point source
        !       Compute the margins in all directions needed for the d/dz operator
        !   Obtain the dimensions in all directions of the source
        !   Initialize a field space including possible margins
        !   Initialize a source space with the primary source definition
        !   Export the source space
        !   Compute the convolution of the source with the Green's function to obtain
        !     the primary field in the field space
        !   Destruct the source space
        !   If the source is an F-source and the aperture is not a point source
        !       Apply the d/dz operator to the field space
        !   Reduce the field space with margins to a field space without margins
        !
        ! =============================================================================

        call SwStartAndCount(cswLinStep); 
        call PrintToLog("PrimarySourcetoField", 1)

        ! 1) Compute the necessary margin (in all directions) to compute the linear step
        !KH not only an extra margin in the z-direction, also in t,x and/or y if the theta's are /= pi/2!!
        !KH this is only necessary if dp/dz needs to be determined, i.e. in case of an F source
        if (cSourceParams%srcshapetype /= iHI_POINTSOURCE .AND. cModelParams.PrimarySourceType == iEI_FSOURCE) then
            aiMargin(1) = ceiling(abs(cModelParams%FDXOrder/2*cSpace%dTanT) - 1e-5)
            aiMargin(2) = ceiling(abs(cModelParams%FDXOrder/2*cSpace%dTanX) - 1e-5)
            aiMargin(3) = ceiling(abs(cModelParams%FDXOrder/2*cSpace%dTanY) - 1e-5)
            aiMargin(4) = cModelParams%FDXOrder/2; 
        else
            aiMargin = (/0, 0, 0, 0/)
        end if

        !New global start positions in Z and T
        iMarginStartZ = cSpace%iStartZ - aiMargin(4)
        iMarginStartT = cSpace%iStartT + iBeamOffsetT(cSpace, -aiMargin(4)) - aiMargin(1)

        ! Determine source dimensions: determined by either the dim+margin of
        ! the field or the minimum size required for the source+margin.
        call SourceDimensions(iSourceDimT, iSourceDimX, iSourceDimY, cSpace%dDt, cSpace%dDx)
        ! The dimension of the T-axis should be equal for the source and field
        ! space. Take the field space size for both.
        iSourceDimT = cSpace%iDimT + 2*aiMargin(1)

        ! 2) Create a space (cSpaceMarg) with the extra margin in all txyz-directions
        call InitSpace(cSpaceMargin, iSI_FIELD, cModelParams.UseYSymmetry, &
                       iSourceDimT, cSpace%iDimX + 2*aiMargin(2), cSpace%iDimY + 2*aiMargin(3), cSpace%iDimZ + 2*aiMargin(4), &
                       iMarginStartT, 0_i8b, 0_i8b, iMarginStartZ, 0_i8b, 0_i8b, 0_i8b, &
                       cSpace%dFnyq, cSpace%dTanX, cSpace%dTanY, cSpace%dTanT)
        call InitGrid(cSpaceMargin, cSpace%cGrid%iProcN, cSpace%cGrid%iProcID, (/.true., .true., .true., .true./)); 
        ! Correct the aiMargins for X and Y, may be changed because of resetting array size in InitGrid
        ! NB This only gives the correct value for the Start margin, which may or may not be the
        ! same as the End margin...
        ! (however, the start margin is the only one that is needed in CopyDataD1/2 anyway...)
        ! the correction needs to be done for the X and Y dimensions only, for the T and Z dimensions,
        ! the array size resetting only gives extra points at the end, start margin is unaffected
        aiMargin(2) = cSpace%iStartX - (cSpaceMargin%iStartX + iBeamOffsetX(cSpaceMargin, aiMargin(4))); 
        aiMargin(3) = cSpace%iStartY - (cSpaceMargin%iStartY + iBeamOffsetY(cSpaceMargin, aiMargin(4))); 

        write (acTemp, '("aiMargin in T, X, Y, Z: ", 4I2)') aiMargin
        call PrintToLog(acTemp, 2)

        ! 2b) Initialize the source
        call InitSpace(cSource, iSI_PRIMARYSOURCE, .false., &
                       iSourceDimT, iSourceDimX, iSourceDimY, 1_i8b, &
                       -aiMargin(1), 0_i8b, 0_i8b, 0_i8b, 0_i8b, 0_i8b, 0_i8b, &
                       cSpaceMargin%dFnyq, cSpaceMargin%dTanX, cSpaceMargin%dTanY, cSpaceMargin%dTanT)
        call InitGrid(cSource, cSpaceMargin%cGrid%iProcN, cSpaceMargin%cGrid%iProcID, (/.true., .true., .true., .true./)); 
        call GridDistr0CreateEmpty(cSource%cGrid); 
        call InitSource(cSource); ! Initialize the source as defined in the config file

        ! Export the source, only if nearfield is obtained
        if (cSpace%iStartZ == nint(cModelParams.Sz/cSpace%dDx)) then
            call ExportSlice(trim(sOutputDir)//trim(sOutputRoot)//"src", &
                             "Source", &
                             cSource, &
                             (/-cSource%iStartT, 0_i8b, 0_i8b, 0_i8b/), &
                             (/cSpace%iDimT, cSource%iDimX, cSource%iDimY, 1_i8b/), &
                             cModelParams.xyzsliceindex, cSource%iDimZ, .true.); 
            if (cSpace%cGrid%iProcID == 0) then
                dTempPhaseDelay = cSourceParams.maxphasedelay/cModelParams%freq0
                call ExportArray(trim(sOutputDir)//trim(sOutputRoot)//"srcdelay", &
                                 "phasedelay", (/dTempPhaseDelay/), cSource%cGrid%iProcID)
            end if
        end if

        ! 3) Compute the convolution on cSpaceMarg
        call PrintToLog('Compute the convolution in the linear step', 2); 
        call ConvolutionGreen(cSource, cSpaceMargin, .false.); 
        call DestructSpace(cSource)

        ! 4) Compute the Finite Difference operator on cSpaceMarg
        ! Only for an F-source excitation type and if the source is not a point source
        if (cSourceParams%srcshapetype /= iHI_POINTSOURCE .AND. cModelParams.PrimarySourceType == iEI_FSOURCE) then
            call PrintToLog('Compute the Finite Difference in z on the field', 2); 
            call BeamDerivativeZ(cSpaceMargin)
        end if

        ! 5a) Strip the d/dz margins in T
        !REMARK: this step is not necessary if there are no margins, i.e. if d/dz was not obtained...
        !                 it would save some time if in that case we simply copy the content of cSpaceXYZMargin to CSpace
        call PrintToLog('Store the results in a beam without margins', 2); 
        call InitSpace(cSpaceXYZMargin, iSI_FIELD, cModelParams.UseYSymmetry, &
                       cSpace%iDimT, cSpaceMargin%iDimX, cSpaceMargin%iDimY, cSpaceMargin%iDimZ, &
                       iMarginStartT + aiMargin(1), 0_i8b, 0_i8b, iMarginStartZ, 0_i8b, 0_i8b, 0_i8b, &
                       cSpace%dFnyq, cSpace%dTanX, cSpace%dTanY, cSpace%dTanT)
        call InitGrid(cSpaceXYZMargin, cSpace%cGrid%iProcN, cSpace%cGrid%iProcID, (/.false., .false., .false., .false./)); 
        call GridDistr0CreateEmpty(cSpaceXYZMargin.cGrid); 
        call CopyDataD1(cSpaceMargin%cGrid, cSpaceXYZMargin.cGrid, aiMargin(1)); 
        call DestructSpace(cSpaceMargin); 
!     call PrintToLog("Variables in cSpaceXYZMargin",-1)
!     call SpaceInfo(cSpaceXYZMargin)

        ! 5b) Strip the d/dz margins in XYZ
        !   - Reorder to distribution 2
        call ReorderDistr0ToDistr1(cSpaceXYZMargin%cGrid); 
        call ReorderDistr1ToDistr2(cSpaceXYZMargin%cGrid); 
        
        call GridDistr2CreateEmpty(cSpace%cGrid); 
        !   - Strip
        call CopyDataD2(cSpaceXYZMargin%cGrid, cSpace%cGrid, aiMargin(2:4)); 
!     call PrintToLog("Variables in cSpace",-1) 
!     call SpaceInfo(cSpace)

        !   - Revert to distribution 0
        call DestructSpace(cSpaceXYZMargin); 

        call ReorderDistr2ToDistr1(cSpace%cGrid); 
        call ReorderDistr1ToDistr0(cSpace%cGrid); 

        cSpace%iSpaceIdentifier = iSI_FIELD; 
        ! Deinitialize
        call SWStop(cswLinStep); 
    END SUBROUTINE PrimarySourcetoField

    SUBROUTINE PlanewaveField(cSpace)

        ! =============================================================================
        !
        !   Programmer: Koos Huijssen
        !
        !   Language: Fortran 90
        !
        !   Version Date    Comment
        !   ------- -----   -------
        !   1.0     090505  Original code (KH)
        !
        ! *****************************************************************************
        !
        !   DESCRIPTION
        !
        !   The subroutine PlanewaveField computes the primary field solution in the
        !   form of a plane wave running through the space in the positive z-direction
        !
        ! *****************************************************************************
        !
        !   INPUT/OUTPUT PARAMETERS
        !
        !   cSpace   io   type(space)  Space structure for which the primary field
        !                              solution is to be obtained
        !
        type(Space), intent(inout)::                cSpace; 
        ! *****************************************************************************
        !
        !   LOCAL PARAMETERS
        !
        !   cSource   type(Space)   Space containing the primary source - is
        !                       mainly used to get the signature right
        !   dZpos        dp     Position in z in wavelengths
        !   dDiffT       dp     Time difference between the source and field space
        !   dOmega       dp     Step size in omega
        !   dCorrection  dpc    Complex correction term to account for the time delay
        !   iLi          i8b    Loop counter over all xyz positions in cSpace
        !   iOmega       i8b    Loop counter over the angular frequency components
        !   acTemp       char   Temporary char array for output log messages
        !
        type(Space) ::                                cSource
        character(len=1024)::       acTemp
        real(dp) ::                                        dZpos, dDiffT, dOmega
        complex(dpc) ::                                dCorrection
        integer(i8b)::              iLi, iOmega
        integer(i8b) ::             iErr

        ! *****************************************************************************
        !
        !   I/O
        !
        !   log file entries
        !
        ! *****************************************************************************
        !
        !   SUBROUTINES/FUNCTIONS CALLED
        !
        !   SwStartAndCount
        !   PrintToLog
        !   GridDistr0CreateEmpty
        !   InitSpace
        !   InitGrid
        !   InitSource_Planewave
        !   ExportSlice
        !   ReorderDistr0ToDistr1
        !   MapVt
        !   TransformT
        !   GridDistr1CreateEmpty
        !   TransformTInv
        !   MapVtInv
        !   ReorderDistr1ToDistr0
        !   DestructSpace
        !   SWStop
        !
        ! *****************************************************************************
        !
        !   PSEUDOCODE
        !
        !   Initialize a source space with the primary source definition - primarily
        !     for the signature
        !   Export the source space
        !   Redistribute, Map in time and Fourier transform the source space in time
        !   For each xyz-position do
        !       Determine the time delay factor for the plane wave solution
        !       Obtain the field on this position as a delayed form of the source
        !         signature
        !   Destruct the source space
        !   Redistribute, Inverse map and inverse Fourier transform the source space in
        !     time
        !
        ! =============================================================================

        call SwStartAndCount(cswLinStep); 
        call PrintToLog("PrimarySourcetoField", 1)

        ! 1) Obtain the source pulse
        ! We need one source trace on each processor.
        call InitSpace(cSource, iSI_PRIMARYSOURCE, .false., &
                       cSpace%iDimT, cSpace%cGrid%iProcN, 1_i8b, 1_i8b, &
                       0_i8b, 0_i8b, 0_i8b, 0_i8b, 0_i8b, 0_i8b, 0_i8b, &
                       cSpace%dFnyq, cSpace%dTanX, cSpace%dTanY, cSpace%dTanT)
        ! The spatial dimensions have no physical meaning, but we need one time frame on each processor
        call InitGrid(cSource, cSpace%cGrid%iProcN, cSpace%cGrid%iProcID, (/.false., .false., .false., .false./)); 
        call GridDistr0CreateEmpty(cSource%cGrid); 
        call InitSource_Planewave(cSource); ! Initialize the source signature as defined in the config file
        ! Configuration is treated as though it is a point source

        if(cSpace%iStartZ==nint(cModelParams.Sz/cSpace%dDx)) then
            call ExportSlice(        trim(sOutputDir) // trim(sOutputRoot) // "src",&
                "Source", &
                cSource, &
                (/ 0_i8b, 0_i8b, 0_i8b, 0_i8b /), &
                (/ cSource%iDimT, cSource%iDimX, cSource%iDimY, 1_i8b /), &
                cModelParams.xyzsliceindex, cSource%iDimZ, .true.);
        end if
        
        ! 2) Transform source pulse in time
        call PrintToLog("Convert the source from distribution 0 to 1", 3)
        call ReorderDistr0ToDistr1(cSource%cGrid); 
        call MapVt(cSource)
        call PrintToLog("Compute the Fourier Transform with respect to the T-axis", 3)
        call TransformT(cSource%cGrid)

        ! 3) Create cSpace grid and fill it with source pulse delayed in z
        ! for each location on this processor:
        !   - get correct delay factor as a function of z
        !        - multiply the source signature with it and put this in the time trace
        call GridDistr1CreateEmpty(cSpace%cGrid); 
        dDiffT = (cSpace%iStartT - cSource%iStartT)*cSpace%dDt
        do iLi = 0, cSpace%cGrid%iD0LocN - 1

            !Z Position
            dZpos = (cSpace%cGrid%aiD0Loc(1 + iLi, 3) + cSpace%iStartZ)*cSpace%dDx
            do iOmega = 0, cSpace%cGrid%iD1TL - 1

                dOmega = two_pi*iOmega*cSpace%dFnyq/cSpace%iDimT

                ! Account for shift in z and in t, and FFT correction factor
                dCorrection = exp(-im*dOmega*(dZpos - dDiffT))/real(2*cSpace%iDimT, dp)

                cSpace%cGrid%pacD1(1 + iLi*cSpace%cGrid%iD1IS + iOmega*cSpace%cGrid%iD1TS) = &
                    cSource%cGrid%pacD1(1 + iOmega*cSpace%cGrid%iD1TS)*dCorrection

            end do

        end do
        call DestructSpace(cSource)

        ! 4) Inverse transform in time by reverting to distr. 0
        call PrintToLog("Compute the inverse Fourier Transform with respect to the T-axis", 3)
        call TransformTInv(cSpace%cGrid)
        call MapVtInv(cSpace)
        call PrintToLog("Redistribute the result to distribution 0", 3)
        call ReorderDistr1ToDistr0(cSpace%cGrid); 
        ! Deinitialize
        call SWStop(cswLinStep); 
    END SUBROUTINE PlanewaveField

    SUBROUTINE PointSourceCloudField(cSpace)

        ! =============================================================================
        !
        !   Programmer: Agisilaos Matalliotakis // Date : 211108
        !
        !   Language: Fortran 90
        !
        !   Version Date    Comment
        !   ------- -----   -------
        !   1.0     090505  Original code (KH)
        !
        ! *****************************************************************************
        !
        !   DESCRIPTION
        !
        !   The subroutine PointSourceCloudField computes the primary field solution
        !   generated from a cloud of point sources, having the same amplitude.
        !
        ! *****************************************************************************
        !
        !   INPUT/OUTPUT PARAMETERS
        !
        !   cSpace   io   type(space)  Space structure for which the primary field
        !                              solution is to be obtained
        !
        type(Space), intent(inout)::                cSpace; 
        ! *****************************************************************************
        !
        !   LOCAL PARAMETERS
        !
        !   acTemp       char   Temporary char array for output log messages
        !
        character(LEN=1024)::                        acTemp; 
        ! *****************************************************************************
        !
        !   I/O
        !
        !   log file entries
        ! =============================================================================
        call PrintToLog("PointSourceCloudField", 1)
        call GridDistr0CreateEmpty(cSpace%cGrid); 
        ! Test whether we have the right input, otherwise quit immediately
        if (cSpace%iSpaceIdentifier /= iSI_FIELD) then
            write (acTemp, '("Error occurred during the execution of PointSourceCloudField, aborting program")'); 
            call PrintToLog(acTemp, -1); 
            write (acTemp, '("the space provided as input is not a FIELD.")'); 
            call PrintToLog(acTemp, -1); 
            stop
        end if

        call SwStartAndCount(cswLinStep); 
        call PointSourceCloudOperator(cSpace)
        cSpace%iSpaceIdentifier = iSI_CONTRASTSOURCE; 
        call test_isnan(cSpace)
        call ContrastSourcetoField(cSpace, cSpace, .false.)

        ! Deinitialize
        call SWStop(cswLinStep); 
    END SUBROUTINE PointSourceCloudField

    SUBROUTINE FieldtoContrastSource(cSpace, iContrastID, cInhomContrast)

        ! =============================================================================
        !
        !   Programmer: Koos Huijssen
        !
        !   Language: Fortran 90
        !
        !   Version Date    Comment
        !   ------- -----   -------
        !   1.0     090505  Original code (KH)
        !
        ! *****************************************************************************
        !
        !   DESCRIPTION
        !
        !   The subroutine FieldToContrastSource computes the contrast source for the
        !   given space - the type of the contrast source is determined by iContrastID.
        !
        ! *****************************************************************************
        !
        !   INPUT/OUTPUT PARAMETERS
        !
        !   cSpace          io  type(Space)   Space structure for which the contrast
        !                                     source is to be obtained
        !   iContrastID     i   i8b           Contrast Identification number
        !   cInhomContrast  i   type(Space)   Space containing the inhomogeneity
        !                                     contrast - if applicable
        !
        type(Space), intent(inout) ::        cSpace
        integer(i8b), intent(in) ::                iContrastID
        type(Space), intent(in) ::                cInhomContrast

        ! *****************************************************************************
        !
        !   LOCAL PARAMETERS
        !
        !   acTemp       char   Temporary char array for output log messages
        !
        character(LEN=1024)::                        acTemp; 
        ! *****************************************************************************
        !
        !   I/O
        !
        !   log file entries
        !
        ! *****************************************************************************
        !
        !   SUBROUTINES/FUNCTIONS CALLED
        !
        !   SwStartAndCount
        !   PrintToLog
        !   NonlinContrastOperator_Ali
        !   InhomContrastOperator_Ali
        !   NonlinContrastOperator
        !   InhomContrastOperator
        !   SWStop
        !
        ! =============================================================================

        call PrintToLog("FieldtoContrastSource", 1); 
        call SwStartAndCount(cswNonLinContr); 
        ! Test whether we have the right input, otherwise quit immediately
        if (cSpace%iSpaceIdentifier /= iSI_FIELD) then
            write (acTemp, '("Error occurred during the execution of BeamCalcNonLinContrast, aborting program")'); 
            call PrintToLog(acTemp, -1); 
            write (acTemp, '("the space provided as input is not a FIELD.")'); 
            call PrintToLog(acTemp, -1); 
            stop
        else if (cSpace%cGrid%iDistr /= 0) then
            write (acTemp, '("Error occurred during the execution of BeamCalcNonLinContrast, aborting program")'); 
            call PrintToLog(acTemp, -1); 
            write (acTemp, '("the grid provided as input is not in distribution 0, but in distribution ", I3, "")') cSpace%cGrid%iDistr; 
            call PrintToLog(acTemp, -1); 
            stop
        end if

        ! Compute the contrast 
        call PrintToLog("Compute the contrast operator", 2); 
        if (cModelParams.UseAntiAliasing) then
            select case (iContrastID)
            case (iCI_NONLINKAPPADX)
                call NonlinearKappaOperatorDXAli(cSpace); 
            case (iCI_NONLINKAPPADY)
                call NonlinearKappaOperatorDYAli(cSpace); 
            case (iCI_NONLINKAPPADZ)
                call NonlinearKappaOperatorDZAli(cSpace); 
            case (iCI_NONLIN)
                call NonlinContrastOperator_Ali(cSpace);
                ! call LagrangianDensity_Ali(cSpace)
                ! call LagrangianDensity_Simpl_Ali(cSpace)
            case (iCI_SCATTERER)                                                                    ! A.M Added 29/01/2020
                call BubbleContrastOperator(cSpace)
            case (iCI_COMPLEXCONTRAST, iCI_SPHERE, iCI_BLOB, iCI_LUNEBERG)
                call InhomContrastOperator_Ali(cSpace); 
            end select
        else
            select case (iContrastID)
            case (iCI_NONLIN) 
                call NonlinContrastOperator(cSpace);
                ! call LagrangianDensity_Ali(cSpace)
                ! call LagrangianDensity_Simpl_Ali(cSpace)
            case (iCI_COMPLEXCONTRAST, iCI_SPHERE, iCI_BLOB, iCI_LUNEBERG)
                call InhomContrastOperator(cSpace, cInhomContrast); 
            case (iCI_SCATTERER)                                                      ! A.M Added 29/01/2020
                call BubbleContrastOperator(cSpace)
            end select
        end if
        cSpace%iSpaceIdentifier = iSI_CONTRASTSOURCE; 
        call SWStop(cswNonLinContr); 
    END SUBROUTINE FieldtoContrastSource

    !!LCSM L.D. 12-07-2010
    SUBROUTINE FieldtoContrastSourceLin(cSpace, IncSpace, iContrastID, cInhomContrast)

        ! =============================================================================
        !
        !   Programmer: Libertario Demi
        !
        !   Language: Fortran 90
        !
        !   Version Date    Comment
        !   ------- -----   -------
        !   1.0     090505  Original code (KH)
        !
        ! *****************************************************************************
        !
        !   DESCRIPTION
        !
        !   The subroutine FieldtoContrastSourceLin computes the contrast source for the
        !   given space - the type of the contrast source is determined by iContrastID.
        !
        ! *****************************************************************************
        !
        !   INPUT/OUTPUT PARAMETERS
        !
        !   cSpace          io  type(Space)   Space structure for which the contrast
        !                                     source is to be obtained
        !   iContrastID     i   i8b           Contrast Identification number
        !   cInhomContrast  i   type(Space)   Space containing the inhomogeneity
        !                                     contrast - if applicable
        !
        type(Space), intent(inout) ::        cSpace, IncSpace
        integer(i8b), intent(in) ::                iContrastID
        type(Space), intent(in) ::                cInhomContrast

        ! *****************************************************************************
        !
        !   LOCAL PARAMETERS
        !
        !   acTemp       char   Temporary char array for output log messages
        !
        character(LEN=1024)::                        acTemp; 
        ! *****************************************************************************
        !
        !   I/O
        !
        !   log file entries
        !
        ! *****************************************************************************
        !
        !   SUBROUTINES/FUNCTIONS CALLED
        !
        !   SwStartAndCount
        !   PrintToLog
        !   NonlinContrastOperator_Ali
        !   InhomContrastOperator_Ali
        !   NonlinContrastOperator
        !   InhomContrastOperator
        !   SWStop
        !
        ! =============================================================================

        call PrintToLog("FieldtoContrastSource", 1); 
        call SwStartAndCount(cswNonLinContr); 
        ! Test whether we have the right input, otherwise quit immediately
        if (cSpace%iSpaceIdentifier /= iSI_FIELD) then

        else if (cSpace%cGrid%iDistr /= 0) then
            write (acTemp, '("Error occurred during the execution of BeamCalcNonLinContrast, aborting program")'); 
            call PrintToLog(acTemp, -1); 
            write (acTemp, '("the grid provided as input is not in distribution 0, but in distribution ", I3, "")') cSpace%cGrid%iDistr; 
            call PrintToLog(acTemp, -1); 
            stop
        end if

        ! Compute the contrast
        call PrintToLog("Compute the contrast operator", 2); 
        if (cModelParams.UseAntiAliasing) then
            select case (iContrastID)
            case (iCI_NONLIN)
                call NonlinContrastOperator_Ali(cSpace); 
            case (iCI_SCATTERER)                                                                   ! A.M Added 29/01/2020
                call BubbleContrastOperator(cSpace)
            !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            case (iCI_LINCONTRA)                                        !! LCSM L.D. 12-07-2010 6
                call NonlinContrastOperatorlin_Ali(cSpace, IncSpace);        !! LCSM L.D. 12-07-2010
            !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            case (iCI_CGCONTRA)                                         !! CG L.D. 12-07-2010   7
                call NonlinContrastOperatorCG_Ali(cSpace);                    !! CG L.D. 12-07-2010
            !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            case (iCI_CGCONVERSION)                                     !! CG L.D. 12-07-2010   8
                call ConversionCG_Ali(cSpace);                                !! CG L.D. 12-07-2010
            !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            case (iCI_CGTHETATWO)                                       !! CG L.D. 12-07-2010   9
                call Theta2CG_Ali(cSpace);                                    !! CG L.D. 12-07-2010
            !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            case (iCI_NONLINKAPPATHETADX)                                 !! NonliKappa L.D. 16-02-2012   13
                call NonlinearKappaOperatorLinDXAli(cSpace, IncSpace);          !! NonliKappa L.D. 16-02-2012
            !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            case (iCI_NONLINKAPPATHETADY)                                 !! NonliKappa L.D. 16-02-2012   14
                call NonlinearKappaOperatorLinDYAli(cSpace, IncSpace);          !! NonliKappa L.D. 16-02-2012
            !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            case (iCI_NONLINKAPPATHETADZ)                                 !! NonliKappa L.D. 16-02-2012   15
                call NonlinearKappaOperatorLinDZAli(cSpace, IncSpace);          !! NonliKappa L.D. 16-02-2012
            !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            case (iCI_NONLINKAPPADIRECTIONDX)                             !!
                call NonlinearKappaOperatorDirectionDXAli(cSpace);              !! Direction part for nonlinear kappa 16
            !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            case (iCI_NONLINKAPPADIRECTIONDY)                             !!
                call NonlinearKappaOperatorDirectionDYAli(cSpace);              !! Direction part for nonlinear kappa 17
            !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            case (iCI_NONLINKAPPADIRECTIONDZ)                             !!
                call NonlinearKappaOperatorDirectionDZAli(cSpace);              !! Direction part for nonlinear kappa 18
            !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            case (iCI_COMPLEXCONTRAST, iCI_SPHERE, iCI_BLOB, iCI_LUNEBERG)
                call InhomContrastOperator_Ali(cSpace); 
            end select
        else
            select case (iContrastID)
            case (iCI_NONLIN)
                call NonlinContrastOperator(cSpace); 
            case (iCI_SCATTERER)                                                   ! A.M Added 29/01/2020
                call BubbleContrastOperator(cSpace)
            case (iCI_COMPLEXCONTRAST, iCI_SPHERE, iCI_BLOB, iCI_LUNEBERG)
                call InhomContrastOperator(cSpace, cInhomContrast); 
            end select
        end if
        cSpace%iSpaceIdentifier = iSI_CONTRASTSOURCE; 
        call SWStop(cswNonLinContr); 
    END SUBROUTINE FieldtoContrastSourceLin

    SUBROUTINE ContrastSourcetoField(cSrcSpace, cDestSpace, inplacemode)

        ! =============================================================================
        !
        !   Programmer: Koos Huijssen
        !
        !   Language: Fortran 90
        !
        !   Version Date    Comment
        !   ------- -----   -------
        !   1.0     090505  Original code (KH)
        !
        ! *****************************************************************************
        !
        !   DESCRIPTION
        !
        !   The subroutine ContrastSourcetoField computes the field in a given field
        !   space as generated by a given contrast source.
        !
        ! *****************************************************************************
        !
        !   INPUT/OUTPUT PARAMETERS
        !
        !   cSrcSpace    io   type(Space)   Space structure containing the contrast
        !                                   source
        !   cDestSpace   io   type(Space)   Space structure to be filled with the
        !                                   field solution
        !   inplacemode   i   lgt           Whether (true) or not (false) to employ an
        !                                   inplace computation of the convolution - is
        !                                   more memory efficient
        !
        type(Space), intent(inout)::            cSrcSpace, cDestSpace
        logical(lgt), intent(in)::                                inplacemode

        ! *****************************************************************************
        !
        !   LOCAL PARAMETERS
        !
        !   acTemp       char   Temporary char array for output log messages
        !
        character(LEN=1024)::                        acTemp; 
        integer(i8b) :: iLi

        ! *****************************************************************************
        !
        !   I/O
        !
        !   log file entries
        !
        ! *****************************************************************************
        !
        !   SUBROUTINES/FUNCTIONS CALLED
        !
        !   SwStartAndCount
        !   PrintToLog
        !   ConvolutionGreen
        !   SWStop  
        !
        ! =============================================================================

        call PrintToLog("ContrastSourcetoField", 1); 
        call SwStartAndCount(cswCorrStep); 
        ! Apply the convolution over it
        call ConvolutionGreen(cSrcSpace, cDestSpace, inplacemode);  
        ! cDestSpace is now the field
        cDestSpace%iSpaceIdentifier = iSI_FIELD

        call SWStop(cswCorrStep); 
    END SUBROUTINE ContrastSourcetoField

    SUBROUTINE ContrastSourcetoFieldConj(cSrcSpace, cDestSpace, inplacemode)

        ! =============================================================================
        !
        !   Programmer: Libertario Demi
        !
        !   Language: Fortran 90
        !
        !   Version Date    Comment
        !   ------- -----   -------
        !   1.0     210512  Original code (LD)
        !
        ! *****************************************************************************
        !
        !   DESCRIPTION
        !
        !   The subroutine ContrastSourcetoField computes the field in a given field
        !   space as generated by a given contrast source.
        !
        ! *****************************************************************************
        !
        !   INPUT/OUTPUT PARAMETERS
        !
        !   cSrcSpace    io   type(Space)   Space structure containing the contrast
        !                                   source
        !   cDestSpace   io   type(Space)   Space structure to be filled with the
        !                                   field solution
        !   inplacemode   i   lgt           Whether (true) or not (false) to employ an
        !                                   inplace computation of the convolution - is
        !                                   more memory efficient
        !
        type(Space), intent(inout)::            cSrcSpace, cDestSpace
        logical(lgt), intent(in)::                                inplacemode

        ! *****************************************************************************
        !
        !   LOCAL PARAMETERS
        !
        !   acTemp       char   Temporary char array for output log messages
        !
        character(LEN=1024)::                        acTemp; 
        ! *****************************************************************************
        !
        !   I/O
        !
        !   log file entries
        !
        ! *****************************************************************************
        !
        !   SUBROUTINES/FUNCTIONS CALLED
        !
        !   SwStartAndCount
        !   PrintToLog
        !   ConvolutionGreen
        !   SWStop
        !
        ! =============================================================================

        call PrintToLog("ContrastSourcetoField", 1); 
        call SwStartAndCount(cswCorrStep); 
        ! Apply the convolution over it
        call ConvolutionGreenConj(cSrcSpace, cDestSpace, inplacemode); 
        ! cDestSpace is now the field
        cDestSpace%iSpaceIdentifier = iSI_FIELD

        call SWStop(cswCorrStep); 
    END SUBROUTINE ContrastSourcetoFieldConj

    SUBROUTINE CFPlaneSourcetoField(cSrcSpace, cDestSpace)

        ! =============================================================================
        !
        !   Programmer: Koos Huijssen
        !
        !   Language: Fortran 90
        !
        !   Version Date    Comment
        !   ------- -----   -------
        !   1.0     090505  Original code (KH)
        !
        ! *****************************************************************************
        !
        !   DESCRIPTION
        !
        !   The subroutine CFPlaneSourcetoField computes the field excited by a plane
        !   on which a pressure is defined in a domain behind the pressure plane.
        !   According to Huygens' principle, we would require a closed surface to
        !   compute the field from the (contrast) sources within the closed surface.
        !   Here, we only take a part of the XY-plane to compute the field of the
        !   sources in the shadow region of the surface.
        !
        ! *****************************************************************************
        !
        !   INPUT/OUTPUT PARAMETERS
        !
        !   cSrcSpace   io   type(space)  Space structure in which the pressure plane
        !                                 is stored. It is assumed to be stored on the
        !                                 first plane in z.
        !   cDestSpace  io   type(space)  Space structure for which the field
        !                                 solution is to be obtained
        !
        type(Space), intent(inout)::                cSrcSpace, cDestSpace; 
        ! *****************************************************************************
        !
        !   LOCAL PARAMETERS
        !
        !   iLi             i8b    loop counter over all xyz-positions
        !   iStart          i8b    start of the xy-positions at z=0
        !   iEnd            i8b    end of the xy-positions at z=0
        !   iLT             i8b    loop counter over all t-instants
        !   iIndS           i8b    index for source array data
        !   iIndD           i8b    index for dest array data
        !   iMarginStartZ   i8b    Start margin in Z for d/dz operation
        !   iMarginStartT   i8b    Start margin in T for d/dz operation
        !   cSrcSpaceMargin type(Space)   Space for the field with margins for d/dz
        !   cSpaceMargin    type(Space)   Space for the field with margins for d/dz
        !   cSpaceXYZMargin type(Space)   Space for the field with margins in XYZ only
        !   aiMargin        i8b    Margins in all four dimensions T/X/Y/Z
        !   dTempPhaseDelay dp     Maximum phase delay in the source
        !   acTemp          char   Temporary char array for output log messages
        !
        integer(i8b)::              iLi, iStart, iEnd, iLT, iIndS, iIndD, iMarginStartZ, iMarginStartT; 
        integer(i8b) ::             iErr
        integer(i4b) ::                                iMPIErr
        type(Space)::               cSrcSpaceMargin, cSpaceMargin, cSpaceXYZMargin
        integer(i8b) ::             aiMargin(4)
        real(dp) ::                                        dTempPhaseDelay
        character(len=1024)::       acTemp

        ! *****************************************************************************
        !
        !   I/O
        !
        !   log file entries
        !
        ! *****************************************************************************
        !
        !   SUBROUTINES/FUNCTIONS CALLED
        !
        !   SwStartAndCount
        !   PrintToLog
        !   iBeamOffsetT
        !   iBeamOffsetX
        !   iBeamOffsetY
        !   InitSpace
        !   InitGrid
        !   GridDistr0CreateEmpty
        !   GridDistr0DeAllocate
        !   ConvolutionGreen
        !   DestructSpace
        !   BeamDerivativeZ
        !   CopyDataD1
        !   CopyDataD2
        !   ReorderDistr0ToDistr1
        !   ReorderDistr1ToDistr2
        !   ReorderDistr2ToDistr1
        !   ReorderDistr1ToDistr0
        !   SWStop
        !
        ! *****************************************************************************
        !
        !   PSEUDOCODE
        !
        !   Compute the margins in all directions needed for the d/dz operator
        !   Initialize a field space including possible margins
        !   Initialize a source space including the margin in time
        !   Copy the input pressure plane in to the source space
        !   Compute the convolution of the source with the Green's function to obtain
        !     the field in the field space
        !   Destruct the source space
        !   Apply the d/dz operator to the field space
        !   Reduce the field space with margins to a field space without margins
        !
        ! =============================================================================

        call SwStartAndCount(cswLinStep); 
        call PrintToLog("CFPlaneSourcetoField", 1)

        !The Correction Field source slice is a pressure slice and therefore
        ! acts as a force source. We need to apply the spatial derivative in z
        ! and therefore need a margin.

        ! 1) Compute the necessary margin (in all directions) to compute the linear step
        aiMargin(1) = ceiling(abs(cModelParams%FDXOrder/2*cDestSpace%dTanT) - 1e-5)
        aiMargin(2) = ceiling(abs(cModelParams%FDXOrder/2*cDestSpace%dTanX) - 1e-5)
        aiMargin(3) = ceiling(abs(cModelParams%FDXOrder/2*cDestSpace%dTanY) - 1e-5)
        aiMargin(4) = cModelParams%FDXOrder/2; 
        !                write(acTemp, '("aiMargin in T, X, Y, Z: ", 4I2)') aiMargin
        !                call PrintToLog(acTemp,2)

        !New global start positions in Z and T
        iMarginStartZ = cDestSpace%iStartZ - aiMargin(4)
        iMarginStartT = cDestSpace%iStartT + iBeamOffsetT(cDestSpace, -aiMargin(4)) - aiMargin(1)

        !Determine source dimensions: determined by either the dim+margin of the field or the minimum size
        ! required for the source+margin

        ! 2) Create a space (cSpaceMarg) with the extra margin in all txyz-directions
        call InitSpace(cSpaceMargin, iSI_FIELD, cModelParams.UseYSymmetry, &
                       cDestSpace%iDimT + 2*aiMargin(1), cDestSpace%iDimX + 2*aiMargin(2), cDestSpace%iDimY + 2*aiMargin(3), cDestSpace%iDimZ + 2*aiMargin(4), &
                       iMarginStartT, 0_i8b, 0_i8b, iMarginStartZ, 0_i8b, 0_i8b, 0_i8b, &
                       cDestSpace%dFnyq, cDestSpace%dTanX, cDestSpace%dTanY, cDestSpace%dTanT)
        call InitGrid(cSpaceMargin, cDestSpace%cGrid%iProcN, cDestSpace%cGrid%iProcID, (/.true., .true., .true., .true./)); 
        ! Correct the iXYTMargins, may be changed because of resetting array size in InitGrid
        ! NB This only gives the correct value for the Start margin, which may or may not be the
        ! same as the End margin...
        ! (however, the start margin is the only one that is needed in CopyDataD1/2 anyway...)
        ! the correction needs to be done for the X and Y dimensions only, for the T and Z dimensions,
        ! the array size resetting only gives extra points at the end, start margin is unaffected
        aiMargin(2) = cDestSpace%iStartX - (cSpaceMargin%iStartX + iBeamOffsetX(cSpaceMargin, aiMargin(4))); 
        aiMargin(3) = cDestSpace%iStartY - (cSpaceMargin%iStartY + iBeamOffsetY(cSpaceMargin, aiMargin(4))); 
        !                call PrintToLog("Variables in cSpaceMargin",-1)
        !                call SpaceInfo(cSpaceMargin)

        write (acTemp, '("aiMargin in T, X, Y, Z: ", 4I2)') aiMargin
        call PrintToLog(acTemp, 2)

        !Create cSrcSpaceMargin to include margin in time dimension
        ! and copy the field slice into it
        ! then deallocate grid in cSrcSpace
        call InitSpace(cSrcSpaceMargin, iSI_FIELDSLICE, cModelParams.UseYSymmetry, &
                       cSpaceMargin%iDimT, cSrcSpace%iDimX, cSrcSpace%iDimY, cSrcSpace%iDimZ, &
                       cSrcSpace%iStartT - aiMargin(1), cSrcSpace%iStartX, cSrcSpace%iStartY, cSrcSpace%iStartZ, 0_i8b, 0_i8b, 0_i8b, &
                       cSrcSpace%dFnyq, cSrcSpace%dTanX, cSrcSpace%dTanY, cSrcSpace%dTanT)
        call InitGrid(cSrcSpaceMargin, cSrcSpace%cGrid%iProcN, cSrcSpace%cGrid%iProcID, (/.false., .false., .false., .false./)); 
        call GridDistr0CreateEmpty(cSrcSpaceMargin%cGrid)

        !call PrintToLog("Variables in cSrcSpaceMargin",-1)
        !call SpaceInfo(cSrcSpaceMargin)

        ! Check whether there are any positions in the FIRST z-plane on this processor
        ! if so, obtain first position of z=0-plane on this processor
        do iStart = 0, cSrcSpace.cGrid.iD0LocN - 1
            if (cSrcSpace.cGrid.aiD0Loc(1 + iStart, 3) == 0) then
                exit
            end if
        end do

        !If there are traces on this processor, copy the traces into
        ! cSrcSpaceMargin (watch the margin in T!!!) and immediately multiply
        ! the pressure with the Fsource correction factor
        ! Fsource = -2*Ptempsource: minus since source term should be - d/dz F(x,t)
        if (iStart < cSrcSpace.cGrid.iD0LocN) then

            !Obtain last position of z=0 plane on this processor
            ! This implementation assumes that the index over the z-dimension
            ! is the slowest one compared to the indices over x and y
            do iEnd = iStart, cSrcSpace.cGrid.iD0LocN - 1
                if (cSrcSpace.cGrid.aiD0Loc(1 + iEnd, 3) /= 0) then
                    exit
                end if
            end do

            do iLi = iStart, iEnd - 1
                do iLT = 0, cSrcSpace%cGrid%iD0TL - 1

                    iIndS = 1 + iLi*cSrcSpace%cGrid%iD0TL + iLT*cSrcSpace%cGrid%iD1TS
                    iIndD = 1 + iLi*cSrcSpaceMargin%cGrid%iD0TL + iLT*cSrcSpaceMargin%cGrid%iD1TS + aiMargin(1)
                    cSrcSpaceMargin%cGrid%parD0(iIndD) = -2.0_dp*cSrcSpace%cGrid%parD0(iIndS); 
                end do
            end do

        end if
        call GridDistr0DeAllocate(cSrcSpace%cGrid)

        ! 3) Compute the convolution on cSpaceMargin
        call PrintToLog('Compute the convolution in the Correction Field plane step', 2); 
        call ConvolutionGreen(cSrcSpaceMargin, cSpaceMargin, .false.); 
        call DestructSpace(cSrcSpaceMargin)

        !call ExportSlice(trim(sOutputDir) // trim(sOutputRoot) // "cSpaceMargin",&
        !        "cSpaceMargin",cSpaceMargin, &
        !        (/ 0_i8b, 0_i8b, 0_i8b, 0_i8b /), &
        !        (/ cSpaceMargin%iDimT, cSpaceMargin%iDimX, 1_i8b, cSpaceMargin%iDimZ /), &
        !        .true.);

        ! 4) Compute the Finite Difference operator on cSpaceMarg
        call PrintToLog('Compute the Finite Difference in z', 2); 
        call BeamDerivativeZ(cSpaceMargin)

        ! 5a) Strip the d/dz margins in T
        call PrintToLog('Store the results in a beam without margins', 2); 
        call InitSpace(cSpaceXYZMargin, iSI_FIELD, cModelParams.UseYSymmetry, &
                       cDestSpace%iDimT, cSpaceMargin%iDimX, cSpaceMargin%iDimY, cSpaceMargin%iDimZ, &
                       iMarginStartT + aiMargin(1), 0_i8b, 0_i8b, iMarginStartZ, 0_i8b, 0_i8b, 0_i8b, &
                       cDestSpace%dFnyq, cDestSpace%dTanX, cDestSpace%dTanY, cDestSpace%dTanT)
        call InitGrid(cSpaceXYZMargin, cDestSpace%cGrid%iProcN, cDestSpace%cGrid%iProcID, (/.false., .false., .false., .false./)); 
        call GridDistr0CreateEmpty(cSpaceXYZMargin.cGrid); 
        call CopyDataD1(cSpaceMargin%cGrid, cSpaceXYZMargin.cGrid, aiMargin(1)); 
        call DestructSpace(cSpaceMargin); 
        ! 5b) Strip the d/dz margins in XYZ
        !   - Reorder to distribution 2
        call ReorderDistr0ToDistr1(cSpaceXYZMargin%cGrid); 
        call ReorderDistr1ToDistr2(cSpaceXYZMargin%cGrid); 
        call GridDistr2CreateEmpty(cDestSpace%cGrid); 
        call CopyDataD2(cSpaceXYZMargin%cGrid, cDestSpace%cGrid, aiMargin(2:4)); 
        !   - revert to distribution 0
        call DestructSpace(cSpaceXYZMargin); 
        call ReorderDistr2ToDistr1(cDestSpace%cGrid); 
        call ReorderDistr1ToDistr0(cDestSpace%cGrid); 
        cDestSpace%iSpaceIdentifier = iSI_FIELD; 
        ! Deinitialize
        call SWStop(cswLinStep); 
    END SUBROUTINE CFPlaneSourcetoField

    SUBROUTINE StoreField(cSpace, acStoredFieldName)

        ! =============================================================================
        !
        !   Programmer: Koos Huijssen
        !
        !   Language: Fortran 90
        !
        !   Version Date    Comment
        !   ------- -----   -------
        !   1.0     090505  Original code (KH)
        !
        ! *****************************************************************************
        !
        !   DESCRIPTION
        !
        !   The subroutine StoreField stores the field in a given space to a temporary
        !   file in the temp dir. Only a Distr. 0 data array is stored, it is assumed
        !   that in a later read all other information is known. The storing is done in
        !   a blockwise fashion.
        !
        ! *****************************************************************************
        !
        !   INPUT/OUTPUT PARAMETERS
        !
        !   cSpace             io   type(space)  Space from which the data array needs
        !                                        to be stored.
        !   acStoredFieldName  io   char         Filename of the temp storage file.
        !
        type(Space), intent(inout)::            cSpace
        character(*), intent(in):: acStoredFieldName

        ! *****************************************************************************
        !
        !   LOCAL PARAMETERS
        !
        !   iErr            i8b    Error number
        !   iLi             i8b    Loop counter over the blocks to save
        !   iLast           i8b    Index of the last full block
        !   iBlockL         i8b    Size of the blocks
        !   acFilename      char   Total filename including path and suffix
        !   acTemp          char   Temporary char array for output log messages
        !
        integer(i8b)::                          iErr
        integer(i8b)::                          iLi, iLast; 
        integer(i8b), parameter::               iBlockL = real(2**20, dp); 
        character(len=1024)::                   acFileName
        character(len=1024)::                   acTemp; 
        ! *****************************************************************************
        !
        !   I/O
        !
        !   log file entries and storing of the data array to temporary file
        !
        ! *****************************************************************************
        !
        !   SUBROUTINES/FUNCTIONS CALLED
        !
        !   SwStartAndCount
        !   PrintToLog
        !   int2str
        !   SWStop
        !
        ! =============================================================================

        call PrintToLog("StoreField", 1)

        if (cSpace%cGrid%iDistr /= 0) then
            write (acTemp, '("Error occurred during the execution of StoreField, aborting program")'); 
            call PrintToLog(acTemp, -1); 
            write (acTemp, '("the grid provided as input is not in distribution 0, but in distribution ", I3, "")') cSpace%cGrid%iDistr; 
            call PrintToLog(acTemp, -1); 
            stop
        end if

        ! Write field to disk
        call SwStartAndCount(cswDisk)
        call PrintToLog('Write P_lin  to disk', 2); 
        acFileName = trim(sOutputDir)//acStoredFieldName//int2str(cSpace%cGrid%iProcID); 
        open (unit=iExportUNIT, file=trim(acFileName), status="REPLACE", form="UNFORMATTED", iostat=iErr)
        if (iErr /= 0) then
            write (acTemp, "('Error in StoreField, file ',A20,'does not exist')") acStoredFieldName
            call PrintToLog(acTemp, -1)
            stop
        end if
        iLast = int(cSpace%cGrid%iD0LocSize/iBlockL); 
        do iLi = 0, iLast - 1
            write (unit=iExportUNIT, iostat=iErr) cSpace%cGrid%parD0(1 + iLi*iBlockL:(iLi + 1)*iBlockL); 
        end do
        write (unit=iExportUNIT, iostat=iErr) cSpace%cGrid%parD0(1 + iLast*iBlockL:cSpace%cGrid%iD0LocSize); 
        close (unit=iExportUNIT); 
        call SWStop(cswDisk)

    END SUBROUTINE StoreField

    SUBROUTINE LoadField(cSpace, acStoredFieldName)

        ! =============================================================================
        !
        !   Programmer: Koos Huijssen
        !
        !   Language: Fortran 90
        !
        !   Version Date    Comment
        !   ------- -----   -------
        !   1.0     090505  Original code (KH)
        !
        ! *****************************************************************************
        !
        !   DESCRIPTION
        !
        !   The subroutine LoadField loads the field from a temporary file in the temp
        !   dir to a given space. The data array is in Distr. 0, it is assumed
        !   that all other information in the space is correct. The storing was done in
        !   a blockwise fashion.
        !
        ! *****************************************************************************
        !
        !   INPUT/OUTPUT PARAMETERS
        !
        !   cSpace             io   type(space)  Space to which the data array needs
        !                                        to be loaded.
        !   acStoredFieldName  io   char         Filename of the temp storage file.
        !
        type(Space), intent(inout)::            cSpace
        character(*), intent(in):: acStoredFieldName

        ! *****************************************************************************
        !
        !   LOCAL PARAMETERS
        !
        !   iErr            i8b    Error number
        !   iLi             i8b    Loop counter over the blocks to load
        !   iLast           i8b    Index of the last full block
        !   iBlockL         i8b    Size of the blocks
        !   parBuffer       dp     Temporary buffer for a block load
        !   acFilename      char   Total filename including path and suffix
        !   acTemp          char   Temporary char array for output log messages
        !
        integer(i8b)::                          iErr
        integer(i8b)::                          iLi, iLast; 
        integer(i8b), parameter::               iBlockL = real(2**20, dp); 
        real(dp), allocatable::                  parBuffer(:); 
        character(len=1024)::                   acFileName
        character(len=1024)::                   acTemp; 
        ! *****************************************************************************
        !
        !   I/O
        !
        !   log file entries and loading of the data array from temporary file
        !
        ! *****************************************************************************
        !
        !   SUBROUTINES/FUNCTIONS CALLED
        !
        !   SwStartAndCount
        !   PrintToLog
        !   int2str
        !   SWStop
        !
        ! =============================================================================

        call PrintToLog("LoadField", 1)

        if (cSpace%cGrid%iDistr /= 0) then
            write (acTemp, '("Error occurred during the execution of LoadField, aborting program")'); 
            call PrintToLog(acTemp, -1); 
            write (acTemp, '("the grid provided as input is not in distribution 0, but in distribution ", I3, "")') cSpace%cGrid%iDistr; 
            call PrintToLog(acTemp, -1); 
            stop
        end if

        ! We read the original values and put them in cSpace
        allocate (parBuffer(iBlockL)); 
        iMemAllocated = iMemAllocated + PRODUCT(SHAPE(parBuffer))*dpS

        call PrintToLog("Buffer allocated", 3)
        call PrintToLog('Read Stored field from disk', 2); 
        call SwStartAndCount(cswDiskAcces)
        acFileName = trim(sOutputDir)//acStoredFieldName//int2str(cSpace%cGrid%iProcID); 
        open (unit=iExportUNIT, file=trim(acFileName), status="OLD", form="UNFORMATTED", iostat=iErr)
        if (iErr /= 0) then
            write (acTemp, "('Error in LoadField, file ',A,' does not exist')") trim(acStoredFieldName)
            call PrintToLog(acTemp, -1)
            stop
        end if

        iLast = int(cSpace%cGrid%iD0LocSize/iBlockL); 
        do iLi = 0, iLast - 1
            read (unit=iExportUNIT, iostat=iErr) parBuffer; 
            cSpace%cGrid%parD0(1 + iLi*iBlockL:(iLi + 1)*iBlockL) = parBuffer
        end do
        read (unit=iExportUNIT, iostat=iErr) parBuffer(1:cSpace%cGrid%iD0LocSize - iLast*iBlockL); 
        cSpace%cGrid%parD0(1 + iLast*iBlockL:cSpace%cGrid%iD0LocSize) = parBuffer(1:cSpace%cGrid%iD0LocSize - iLast*iBlockL)
        close (unit=iExportUNIT); 
        call SWStop(cswDiskAcces)

        call SWStop(cswDisk)
        iMemAllocated = iMemAllocated - PRODUCT(SHAPE(parBuffer))*dpS

        deallocate (parBuffer); 
        call PrintToLog("Free buffer", 3)

    END SUBROUTINE LoadField

    SUBROUTINE AddStoredtoField(cSpace, acStoredFieldName)

        ! =============================================================================
        !
        !   Programmer: Koos Huijssen
        !
        !   Language: Fortran 90
        !
        !   Version Date    Comment
        !   ------- -----   -------
        !   1.0     090505  Original code (KH)
        !
        ! *****************************************************************************
        !
        !   DESCRIPTION
        !
        !   The subroutine AddStoredtoField loads the field from a temporary file in
        !   the temp dir and adds it to a given space. The data array is in Distr. 0,
        !   it is assumed that the dimensions of the space is equal to that of the data
        !   array in the file. The storing was done in a blockwise fashion.
        !
        ! *****************************************************************************
        !
        !   INPUT/OUTPUT PARAMETERS
        !
        !   cSpace             io   type(space)  Space to which the stored data array
        !                                        is added.
        !   acStoredFieldName  io   char         Filename of the temp storage file.
        !
        type(Space), intent(inout)::            cSpace
        character(*), intent(in):: acStoredFieldName

        ! *****************************************************************************
        !
        !   LOCAL PARAMETERS
        !
        !   iErr            i8b    Error number
        !   iLi             i8b    Loop counter over the blocks to load
        !   iLast           i8b    Index of the last full block
        !   iBlockL         i8b    Size of the blocks
        !   parBuffer       dp     Temporary buffer for a block load
        !   acFilename      char   Total filename including path and suffix
        !   acTemp          char   Temporary char array for output log messages
        !
        integer(i8b)::                          iErr
        integer(i8b)::                          iLi, iLast; 
        integer(i8b), parameter::               iBlockL = real(2**20, dp); 
        real(dp), allocatable::                 parBuffer(:)
        character(len=1024)::                   acFileName
        character(len=1024)::                   acTemp; 
        ! *****************************************************************************
        !
        !   I/O
        !
        !   log file entries and loading of the data array from temporary file
        !
        ! *****************************************************************************
        !
        !   SUBROUTINES/FUNCTIONS CALLED
        !
        !   SwStartAndCount
        !   PrintToLog
        !   int2str
        !   SWStop
        !
        ! =============================================================================

        call PrintToLog("AddStoredtoField", 1)

        if (cSpace%cGrid%iDistr /= 0) then
            write (acTemp, '("Error occurred during the execution of AddStoredtoField, aborting program")'); 
            call PrintToLog(acTemp, -1); 
            write (acTemp, '("the grid provided as input is not in distribution 0, but in distribution ", I3, "")') cSpace%cGrid%iDistr; 
            call PrintToLog(acTemp, -1); 
            stop
        end if

        ! We read the original values and add them to cSpace
        allocate (parBuffer(iBlockL)); 
        iMemAllocated = iMemAllocated + PRODUCT(SHAPE(parBuffer))*dpS

        call PrintToLog("Buffer allocated", 3)
        call PrintToLog('Read P_lin from disk', 2); 
        call SwStartAndCount(cswDiskAcces)
        acFileName = trim(sOutputDir)//acStoredFieldName//int2str(cSpace%cGrid%iProcID); 
        open (unit=iExportUNIT, file=trim(acFileName), status="OLD", form="UNFORMATTED", iostat=iErr)
        if (iErr /= 0) then
            write (acTemp, "('Error in AddStoredToField, file ',A,' does not exist')") trim(acStoredFieldName)
            call PrintToLog(acTemp, -1)
            stop
        end if

        iLast = int(cSpace%cGrid%iD0LocSize/iBlockL); 
        do iLi = 0, iLast - 1
            read (unit=iExportUNIT, iostat=iErr) parBuffer; 
            cSpace%cGrid%parD0(1 + iLi*iBlockL:(iLi + 1)*iBlockL) = cSpace%cGrid%parD0(1 + iLi*iBlockL:(iLi + 1)*iBlockL) + parBuffer
        end do
        read (unit=iExportUNIT, iostat=iErr) parBuffer(1:cSpace%cGrid%iD0LocSize - iLast*iBlockL); 
        cSpace%cGrid%parD0(1 + iLast*iBlockL:cSpace%cGrid%iD0LocSize) = cSpace%cGrid%parD0(1 + iLast*iBlockL:cSpace%cGrid%iD0LocSize) + parBuffer(1:cSpace%cGrid%iD0LocSize - iLast*iBlockL)
        close (unit=iExportUNIT); 
        call SWStop(cswDiskAcces)

        call SWStop(cswDisk)

        iMemAllocated = iMemAllocated - PRODUCT(SHAPE(parBuffer))*dpS
        deallocate (parBuffer); 
        call PrintToLog("Free buffer", 3)

    END SUBROUTINE AddStoredtoField

    !--------------------------------------------------L.D. 05-01-2011 DFM option
    SUBROUTINE AddStoredtoFieldDFM(A, B)

        type(Space), intent(inout)::            A
        type(Space), intent(in)   ::            B

        integer(i8b)              ::            iDimT, iDimX, iDimY, iDimZ, iD0LocN, SHIFT, iindex, iiindex, istart

        iDimT = A%iDimT
        iDimZ = A%iDimZ
        iDimX = A%iDimX
        iDimY = A%iDimY
        iD0LocN = A%cGrid%iD0LocN
        !print *, idimt,"iDimT"
        !print *, idimx,"iDimX"
        !print *, idimy,"iDimY"
        !print *, idimz,"iDimZ"
        !print *, size(A%cGrid%parD0),"Total size A"
        !print *, size(A%cGrid%parD0),"Total size B"
        !print *, iD0LocN, "iD0LocN"

        !SHIFT=4
        SHIFT = cSourceParams.SHIFT

        do iiindex = 0, (iD0LocN/iDimX) - 1

            istart = iiindex*iDimX*iDimT

            do iindex = 0, iDimX - SHIFT - 1

                A%cGrid%parD0(istart + 1:istart + idimt) = A%cGrid%parD0(istart + 1:istart + idimt) + B%cGrid%parD0(idimt*SHIFT + istart + 1:idimt*SHIFT + istart + idimt)
                istart = istart + idimt
            end do

        end do

    END SUBROUTINE AddStoredtoFieldDFM
    !--------------------------------------------------

    !--------------------------------------------------L.D. 17-09-2010
    SUBROUTINE SubtractData(A, B, C)

        ! =============================================================================
        !
        !   Programmer: Koos Huijssen
        !
        !   Language: Fortran 90
        !
        !   Version Date    Comment
        !   ------- -----   -------
        !   1.0     090505  Original code (KH)
        !
        ! *****************************************************************************
        !
        !   DESCRIPTION
        !
        !   The subroutine SubtractStoredtoField loads the field from a temporary file in
        !   the temp dir and subtract it to a given space. The data array is in Distr. 0,
        !   it is assumed that the dimensions of the space is equal to that of the data
        !   array in the file. The storing was done in a blockwise fashion.
        !
        ! *****************************************************************************
        !
        !   INPUT/OUTPUT PARAMETERS
        !
        !   cSpace             io   type(space)  Space to which the stored data array
        !                                        is subtracted.
        !   acStoredFieldName  io   char         Filename of the temp storage file.
        !
        type(Space), intent(in)   ::            B, C
        type(Space), intent(inout)   ::         A
        integer(i8b)::                          iLi, Vectordimensions
        character(len=1024)::                   acTemp; 
        ! *****************************************************************************
        !
        !   I/O
        !
        !   log file entries and loading of the data array from temporary file
        !
        ! *****************************************************************************
        !
        !   SUBROUTINES/FUNCTIONS CALLED
        !
        !   SwStartAndCount
        !   PrintToLog
        !   int2str
        !   SWStop
        !
        ! =============================================================================

        call PrintToLog("SumSquare", 1)

        if (A%cGrid%iDistr /= 0) then
            write (acTemp, '("Error occurred during the execution of Copy, aborting program")'); 
            call PrintToLog(acTemp, -1); 
            write (acTemp, '("the grid provided as input is not in distribution 0, but in distribution ", I3, "")') A%cGrid%iDistr; 
            call PrintToLog(acTemp, -1); 
            stop
        end if

        Vectordimensions = size(A%cGrid%parD0)

        do iLi = 0, Vectordimensions - 1

            A%cGrid%parD0(1 + iLi) = B%cGrid%parD0(1 + iLi) - C%cGrid%parD0(1 + iLi)!-C%cGrid%parD0(1+iLi)
        end do

        call PrintToLog("Copy Finished", 3)

    END SUBROUTINE SubtractData
    !--------------------------------------------------
    ! L.D. 17-09-2010
    SUBROUTINE sumsquare(realvalue, cSpace, cSpace2)

        ! =============================================================================
        !
        !   Programmer: Libertario Demi
        !
        !   Language: Fortran 90
        !
        !   Version Date    Comment
        !   ------- -----   -------
        !   1.0     210512  Original code (LD)
        !
        ! *****************************************************************************
        !
        !   DESCRIPTION
        ! *****************************************************************************
        !
        !   INPUT/OUTPUT PARAMETERS
        !
        !   realvalue             io   type(real(dp)).
        !   acStoredFieldName     io   char         Filename of the temp storage file.
        !
        real(8), intent(inout):: realvalue

        type(Space), intent(in)   :: cSpace, cSpace2

        ! *****************************************************************************
        !
        !   LOCAL PARAMETERS
        !
        !   iErr            i8b    Error number
        !   iLi             i8b    Loop counter over the blocks to load
        !   iLast           i8b    Index of the last full block
        !   iBlockL         i8b    Size of the blocks
        !   parBuffer       dp     Temporary buffer for a block load
        !   acFilename      char   Total filename including path and suffix
        !   acTemp          char   Temporary char array for output log messages
        !
        integer(i8b)::                          iErr
        integer(i8b)::                          iLi, Vectordimensions
        character(len=1024)::                   acFileName
        character(len=1024)::                   acTemp; 
        ! *****************************************************************************
        !
        !   I/O
        !
        !   log file entries and loading of the data array from temporary file
        !
        ! *****************************************************************************
        !
        !   SUBROUTINES/FUNCTIONS CALLED
        !
        !   SwStartAndCount
        !   PrintToLog
        !   int2str
        !   SWStop
        !
        ! =============================================================================

        call PrintToLog("SumSquare", 1)

        if (cSpace%cGrid%iDistr /= 0) then
            write (acTemp, '("Error occurred during the execution of Sumsquare, aborting program")'); 
            call PrintToLog(acTemp, -1); 
            write (acTemp, '("the grid provided as input is not in distribution 0, but in distribution ", I3, "")') cSpace%cGrid%iDistr; 
            call PrintToLog(acTemp, -1); 
            stop
        end if

        Vectordimensions = size(cSpace%cGrid%parD0)
        realvalue = 0
        do iLi = 0, Vectordimensions - 1

            realvalue = (cSpace%cGrid%parD0(1 + iLi))*(cSpace2%cGrid%parD0(1 + iLi)) + realvalue
        end do

        call PrintToLog("Sum Finished", 3)

    END SUBROUTINE sumsquare

    SUBROUTINE sumandmultiply(A, B, C, D, E, F)
        !L.D. 20-09-2010
        ! =============================================================================
        !
        !   Programmer: Libertario Demi
        !
        !   Language: Fortran 90
        !
        !   Version Date    Comment
        !   ------- -----   -------
        !   1.0     090505  Original code (KH)
        !
        ! *****************************************************************************
        !
        !   DESCRIPTION
        ! *****************************************************************************
        !
        !   INPUT/OUTPUT PARAMETERS
        !
        !   realvalue             io   type(real(dp)).
        !   acStoredFieldName     io   char         Filename of the temp storage file.
        !
        real(8), intent(in):: D, E
        integer(i8b), intent(in)   :: F
        integer(i8b)     :: Vectordimensions, iLi
        type(Space), intent(in)   :: B, C
        type(Space), intent(inout)   :: A
        character(len=1024)::                   acTemp; 
        !
        !   I/O
        !
        !   log file entries and loading of the data array from temporary file
        !
        ! *****************************************************************************
        !
        !   SUBROUTINES/FUNCTIONS CALLED
        !
        !   SwStartAndCount
        !   PrintToLog
        !   int2str
        !   SWStop
        !
        ! =============================================================================

        call PrintToLog("Sum&Multiply", 1)

        if (A%cGrid%iDistr /= 0) then
            write (acTemp, '("Error occurred during the execution of Sumsquare, aborting program")'); 
            call PrintToLog(acTemp, -1); 
            write (acTemp, '("the grid provided as input is not in distribution 0, but in distribution ", I3, "")') A%cGrid%iDistr; 
            call PrintToLog(acTemp, -1); 
            stop
        end if

        Vectordimensions = size(A%cGrid%parD0)
        !print *, "Iteration Number in the sum and multiply function",F

        do iLi = 0, Vectordimensions - 1
            if (F == 1) then
                A%cGrid%parD0(1 + iLi) = B%cGrid%parD0(1 + iLi)
            else
                A%cGrid%parD0(1 + iLi) = B%cGrid%parD0(1 + iLi) + D*(A%cGrid%parD0(1 + iLi) - E*(C%cGrid%parD0(1 + iLi)))
            end if
        end do

        call PrintToLog("Sum Finished", 3)

    END SUBROUTINE sumandmultiply

    SUBROUTINE sumandmultiplyCG(A, b, C)
        !L.D. 20-07-2011
        ! =============================================================================
        !
        !   Programmer: Libertario Demi
        !
        !   Language: Fortran 90
        !
        !   Version Date    Comment
        !   ------- -----   -------
        !   1.0     210512  Original code (LD)
        !
        ! *****************************************************************************
        !
        !   DESCRIPTION
        ! *****************************************************************************
        !
        !   INPUT/OUTPUT PARAMETERS
        !
        !   realvalue             io   type(real(dp)).
        !   acStoredFieldName     io   char         Filename of the temp storage file.
        !
        real(8), intent(in):: b
        integer(i8b)     :: Vectordimensions, iLi
        type(Space), intent(in)   :: C
        type(Space), intent(inout)   :: A
        character(len=1024)::                   acTemp; 
        !
        !   I/O
        !
        !   log file entries and loading of the data array from temporary file
        !
        ! *****************************************************************************
        !
        !   SUBROUTINES/FUNCTIONS CALLED
        !
        !   SwStartAndCount
        !   PrintToLog
        !   int2str
        !   SWStop
        !
        ! =============================================================================

        call PrintToLog("Sum&MultiplyCG", 1)

        if (A%cGrid%iDistr /= 0) then
            write (acTemp, '("Error occurred during the execution of Sumsquare, aborting program")'); 
            call PrintToLog(acTemp, -1); 
            write (acTemp, '("the grid provided as input is not in distribution 0, but in distribution ", I3, "")') A%cGrid%iDistr; 
            call PrintToLog(acTemp, -1); 
            stop
        end if

        Vectordimensions = size(A%cGrid%parD0)
        !print *, "Iteration Number in the sum and multiply function",F

        do iLi = 0, Vectordimensions - 1

            A%cGrid%parD0(1 + iLi) = A%cGrid%parD0(1 + iLi) + b*(C%cGrid%parD0(1 + iLi))

        end do

        call PrintToLog("Sum Finished", 3)

    END SUBROUTINE sumandmultiplyCG

    SUBROUTINE sumandmultiplyCGbis(A, B, C, D)
        !L.D. 20-07-2011
        ! =============================================================================
        !
        !   Programmer: Libertario Demi
        !
        !   Language: Fortran 90
        !
        !   Version Date    Comment
        !   ------- -----   -------
        !   1.0     090505  Original code (KH)
        !
        ! *****************************************************************************
        !
        !   DESCRIPTION
        ! *****************************************************************************
        !
        !   INPUT/OUTPUT PARAMETERS
        !
        !   realvalue             io   type(real(dp)).
        !   acStoredFieldName     io   char         Filename of the temp storage file.
        !

        integer(i8b)     :: Vectordimensions, iLi
        type(Space), intent(in)   :: B, C, D
        type(Space), intent(inout)   :: A
        character(len=1024)::                   acTemp; 
        !
        !   I/O
        !
        !   log file entries and loading of the data array from temporary file
        !
        ! *****************************************************************************
        !
        !   SUBROUTINES/FUNCTIONS CALLED
        !
        !   SwStartAndCount
        !   PrintToLog
        !   int2str
        !   SWStop
        !
        ! =============================================================================

        call PrintToLog("Sum&MultiplyCGbis", 1)

        if (A%cGrid%iDistr /= 0) then
            write (acTemp, '("Error occurred during the execution of Sumsquare, aborting program")'); 
            call PrintToLog(acTemp, -1); 
            write (acTemp, '("the grid provided as input is not in distribution 0, but in distribution ", I3, "")') A%cGrid%iDistr; 
            call PrintToLog(acTemp, -1); 
            stop
        end if

        Vectordimensions = size(A%cGrid%parD0)
        !print *, "Iteration Number in the sum and multiply function",F

        do iLi = 0, Vectordimensions - 1

            A%cGrid%parD0(1 + iLi) = B%cGrid%parD0(1 + iLi) - C%cGrid%parD0(1 + iLi) + D%cGrid%parD0(1 + iLi)

        end do

        call PrintToLog("Sum Finished", 3)

    END SUBROUTINE sumandmultiplyCGbis

    !============================================ L.D. 10-08-2011
    SUBROUTINE sumandmultiplyCGtris(A, B, C)
        !L.D. 20-07-2011
        ! =============================================================================
        !
        !   Programmer: Libertario Demi
        !
        !   Language: Fortran 90
        !
        !   Version Date    Comment
        !   ------- -----   -------
        !   1.0     21-05-12  Original code (LD)
        !
        ! *****************************************************************************
        !
        !   DESCRIPTION
        !   A=A*B-C
        ! *****************************************************************************
        !
        !   INPUT/OUTPUT PARAMETERS
        !
        !   realvalue             io   type(real(dp)).
        !   acStoredFieldName     io   char         Filename of the temp storage file.
        !

        integer(i8b)     :: Vectordimensions, iLi
        type(Space), intent(in)   :: B, C
        type(Space), intent(inout)   :: A
        character(len=1024)::                   acTemp; 
        !
        !   I/O
        !
        !   log file entries and loading of the data array from temporary file
        !
        ! *****************************************************************************
        !
        !   SUBROUTINES/FUNCTIONS CALLED
        !
        !   SwStartAndCount
        !   PrintToLog
        !   int2str
        !   SWStop
        !
        ! =============================================================================

        call PrintToLog("Sum&MultiplyCGtris", 1)

        if (A%cGrid%iDistr /= 0) then
            write (acTemp, '("Error occurred during the execution of SumCGtris, aborting program")'); 
            call PrintToLog(acTemp, -1); 
            write (acTemp, '("the grid provided as input is not in distribution 0, but in distribution ", I3, "")') A%cGrid%iDistr; 
            call PrintToLog(acTemp, -1); 
            stop
        end if

        if (B%cGrid%iDistr /= 0) then
            write (acTemp, '("Error occurred during the execution of SumCGtris, aborting program")'); 
            call PrintToLog(acTemp, -1); 
            write (acTemp, '("the grid provided as input is not in distribution 0, but in distribution ", I3, "")') A%cGrid%iDistr; 
            call PrintToLog(acTemp, -1); 
            stop
        end if

        if (C%cGrid%iDistr /= 0) then
            write (acTemp, '("Error occurred during the execution of SumCGtris, aborting program")'); 
            call PrintToLog(acTemp, -1); 
            write (acTemp, '("the grid provided as input is not in distribution 0, but in distribution ", I3, "")') A%cGrid%iDistr; 
            call PrintToLog(acTemp, -1); 
            stop
        end if

        call PrintToLog("Sum&MultiplyCGtris---------Step1--------", 1)

        Vectordimensions = size(A%cGrid%parD0)
        !print *, "Iteration Number in the sum and multiply function",F
        call PrintToLog("Sum&MultiplyCGtris---------Step2--------", 1)

        do iLi = 0, Vectordimensions - 1
            !call PrintToLog("Sum&MultiplyCGtris---------Step3--------",1)
            !print *, "From matrix A", A%cGrid%parD0(1+iLi)
            !print *, "From matrix B", B%cGrid%parD0(1+iLi)
            !print *, "From matrix C", C%cGrid%parD0(1+iLi)
            A%cGrid%parD0(1 + iLi) = (A%cGrid%parD0(1 + iLi)*B%cGrid%parD0(1 + iLi)) - C%cGrid%parD0(1 + iLi)
            !call PrintToLog("Sum&MultiplyCGtris---------Step4--------",1)
            !A%cGrid%parD0(1+iLi)= conjg((A%cGrid%parD0(1+iLi))*(B%cGrid%parD0(1+iLi))-C%cGrid%parD0(1+iLi))

        end do

        call PrintToLog("SumCGtris Finished", 3)

    END SUBROUTINE sumandmultiplyCGtris
    !============================================ L.D. 10-08-2011

    SUBROUTINE SubtractStoredToFieldBis(A, B, C, D)
        ! =============================================================================
        !
        !   Programmer: Libertario Demi

        ! *****************************************************************************
        !
        !   INPUT/OUTPUT PARAMETERS
        !
        !   A,B,C             io   type(space)  Space to which the stored data array
        !                                        is subtracted.
        !
        type(Space), intent(inout)::            A
        type(Space), intent(in)   ::            B, C
        integer(i8b)::                          iLi, Vectordimensions
        character(len=1024)::                   acTemp; 
        real(8), intent(in):: D

        ! *****************************************************************************
        !
        !   I/O
        !
        !   log file entries and loading of the data array from temporary file
        !
        ! *****************************************************************************
        !
        !   SUBROUTINES/FUNCTIONS CALLED
        !
        !   SwStartAndCount
        !   PrintToLog
        !   int2str
        !   SWStop
        !
        ! =============================================================================

        call PrintToLog("Subtraction and Multiplication", 1)

        if (A%cGrid%iDistr == 2) then
            call ReorderDistr2ToDistr1(A%cGrid)
        end if
        if (A%cGrid%iDistr == 1) then
            call ReorderDistr1ToDistr0(A%cGrid)
        end if

        if (A%cGrid%iDistr /= 0) then
            write (acTemp, '("Error occurred during the execution of Subtraction, aborting program")'); 
            call PrintToLog(acTemp, -1); 
            write (acTemp, '("the grid provided as input is not in distribution 0, but in distribution ", I3, "")') A%cGrid%iDistr; 
            call PrintToLog(acTemp, -1); 
            stop
        end if

        Vectordimensions = size(A%cGrid%parD0)

        do iLi = 0, Vectordimensions - 1

            A%cGrid%parD0(1 + iLi) = B%cGrid%parD0(1 + iLi) - D*C%cGrid%parD0(1 + iLi)
        end do

        call PrintToLog("Subtraction Finished", 3)

    END SUBROUTINE SubtractStoredToFieldBis

    SUBROUTINE SubtractStoredToFieldTris(A, B, C, D, E)
        ! =============================================================================
        !
        !   Programmer: Koos Huijssen
        !
        !   Language: Fortran 90
        !
        !   Version Date    Comment
        !   ------- -----   -------
        !   1.0     090505  Original code (KH)
        !
        ! *****************************************************************************
        !
        !   DESCRIPTION
        !
        !   The subroutine SubtractStoredtoField loads the field from a temporary file in
        !   the temp dir and subtract it to a given space. The data array is in Distr. 0,
        !   it is assumed that the dimensions of the space is equal to that of the data
        !   array in the file. The storing was done in a blockwise fashion.
        !
        ! *****************************************************************************
        !
        !   INPUT/OUTPUT PARAMETERS
        !
        !   cSpace             io   type(space)  Space to which the stored data array
        !                                        is subtracted.
        !   acStoredFieldName  io   char         Filename of the temp storage file.
        !
        type(Space), intent(inout)::            A
        type(Space), intent(in)   ::            B, C
        integer(i8b)::                          iLi, Vectordimensions
        character(len=1024)::                   acTemp; 
        real(8), intent(in):: D, E

        ! *****************************************************************************
        !
        !   I/O
        !
        !   log file entries and loading of the data array from temporary file
        !
        ! *****************************************************************************
        !
        !   SUBROUTINES/FUNCTIONS CALLED
        !
        !   SwStartAndCount
        !   PrintToLog
        !   int2str
        !   SWStop
        !
        ! =============================================================================

        call PrintToLog("SumSquare", 1)

        if (A%cGrid%iDistr /= 0) then
            write (acTemp, '("Error occurred during the execution of Sumsquare, aborting program")'); 
            call PrintToLog(acTemp, -1); 
            write (acTemp, '("the grid provided as input is not in distribution 0, but in distribution ", I3, "")') A%cGrid%iDistr; 
            call PrintToLog(acTemp, -1); 
            stop
        end if

        Vectordimensions = size(A%cGrid%parD0)

        do iLi = 0, Vectordimensions - 1

            A%cGrid%parD0(1 + iLi) = A%cGrid%parD0(1 + iLi) + D*B%cGrid%parD0(1 + iLi) + E*C%cGrid%parD0(1 + iLi)
        end do

        call PrintToLog("Sum Finished", 3)

    END SUBROUTINE SubtractStoredToFieldTris

    SUBROUTINE Copydata(A, B)
        ! =============================================================================
        !
        !   Programmer: Libertario Demi
        !
        !   Language: Fortran 90
        !
        !   Version Date    Comment
        !   ------- -----   -------
        !   1.0     210512  Original code (LD)
        !
        ! *****************************************************************************
        !
        !
        ! *****************************************************************************
        !
        !   INPUT/OUTPUT PARAMETERS
        !
        !   cSpace             io   type(space)  Space to which the stored data array
        !                                        is subtracted.
        !   acStoredFieldName  io   char         Filename of the temp storage file.
        !
        type(Space), intent(in)   ::            B
        type(Space), intent(inout)   ::         A
        integer(i8b)::                          iLi, Vectordimensions
        character(len=1024)::                   acTemp; 
        ! *****************************************************************************
        !
        !   I/O
        !
        !   log file entries and loading of the data array from temporary file
        !
        ! *****************************************************************************
        !
        !   SUBROUTINES/FUNCTIONS CALLED
        !
        !   SwStartAndCount
        !   PrintToLog
        !   int2str
        !   SWStop
        !
        ! =============================================================================

        call PrintToLog("SumSquare", 1)

        if (B%cGrid%iDistr /= 0) then
            write (acTemp, '("Error occurred during the execution of Copy, aborting program")'); 
            call PrintToLog(acTemp, -1); 
            write (acTemp, '("the grid provided as input is not in distribution 0, but in distribution ", I3, "")') B%cGrid%iDistr; 
            call PrintToLog(acTemp, -1); 
            stop
        end if

        Vectordimensions = size(B%cGrid%parD0)

        do iLi = 0, Vectordimensions - 1

            A%cGrid%parD0(1 + iLi) = B%cGrid%parD0(1 + iLi)
        end do

        call PrintToLog("Copy Finished", 3)

    END SUBROUTINE Copydata

    SUBROUTINE SumData(A, B)
        ! =============================================================================
        !
        !   Programmer: Koos Huijssen
        !
        !   Language: Fortran 90
        !
        !   Version Date    Comment
        !   ------- -----   -------
        !   1.0     090505  Original code (KH)
        !
        ! *****************************************************************************
        !
        !   DESCRIPTION
        !
        !   The subroutine SubtractStoredtoField loads the field from a temporary file in
        !   the temp dir and subtract it to a given space. The data array is in Distr. 0,
        !   it is assumed that the dimensions of the space is equal to that of the data
        !   array in the file. The storing was done in a blockwise fashion.
        !
        ! *****************************************************************************
        !
        !   INPUT/OUTPUT PARAMETERS
        !
        !   cSpace             io   type(space)  Space to which the stored data array
        !                                        is subtracted.
        !   acStoredFieldName  io   char         Filename of the temp storage file.
        !
        type(Space), intent(in)   ::            B
        type(Space), intent(inout)   ::         A
        integer(i8b)::                          iLi, Vectordimensions
        character(len=1024)::                   acTemp; 
        ! *****************************************************************************
        !
        !   I/O
        !
        !   log file entries and loading of the data array from temporary file
        !
        ! *****************************************************************************
        !
        !   SUBROUTINES/FUNCTIONS CALLED
        !
        !   SwStartAndCount
        !   PrintToLog
        !   int2str
        !   SWStop
        !
        ! =============================================================================

        call PrintToLog("SumSquare", 1)

        if (B%cGrid%iDistr /= 0) then
            write (acTemp, '("Error occurred during the execution of Copy, aborting program")'); 
            call PrintToLog(acTemp, -1); 
            write (acTemp, '("the grid provided as input is not in distribution 0, but in distribution ", I3, "")') B%cGrid%iDistr; 
            call PrintToLog(acTemp, -1); 
            stop
        end if

        Vectordimensions = size(B%cGrid%parD0)

        do iLi = 0, Vectordimensions - 1

            A%cGrid%parD0(1 + iLi) = A%cGrid%parD0(1 + iLi) + B%cGrid%parD0(1 + iLi)
        end do

        call PrintToLog("Copy Finished", 3)

    END SUBROUTINE SumData

    !------------------------------------------------
    SUBROUTINE StorePlane(cSpace, acStoredFieldName)

        ! =============================================================================
        !
        !   Programmer: Koos Huijssen
        !
        !   Language: Fortran 90
        !
        !   Version Date    Comment
        !   ------- -----   -------
        !   1.0     090505  Original code (KH)
        !
        ! *****************************************************************************
        !
        !   DESCRIPTION
        !
        !   The subroutine StorePlane stores the first plane in Z of the field in a
        !   given space to a temporary file in the temp dir. Only a Distr. 0 data array
        !   is stored, it is assumed that in a later read all other information is
        !   known. The storing is done in a blockwise fashion.
        !
        ! *****************************************************************************
        !
        !   INPUT/OUTPUT PARAMETERS
        !
        !   cSpace             io   type(space)  Space from which the data at the plane
        !                                        needs to be stored.
        !   acStoredFieldName  io   char         Filename of the temp storage file.
        !
        type(Space), intent(inout)::            cSpace
        character(*), intent(in):: acStoredFieldName

        ! *****************************************************************************
        !
        !   LOCAL PARAMETERS
        !
        !   iErr            i8b    Error number
        !   iStart          i8b    Start index of the plane
        !   iLength         i8b    Length of the data array of the plane
        !   iLi             i8b    Loop counter
        !   iLi2            i8b    Loop counter
        !   iLast           i8b    Index of the last full block
        !   iBlockL         i8b    Size of the blocks
        !   acFilename      char   Total filename including path and suffix
        !   acTemp          char   Temporary char array for output log messages
        !
        integer(i8b)::                          iErr, iStart, iLength
        integer(i8b)::                          iLi, iLi2, iLast; 
        integer(i8b), parameter::               iBlockL = 2**48; 
        character(len=1024)::                   acFileName
        character(len=1024)::                   acTemp; 
        ! *****************************************************************************
        !
        !   I/O
        !
        !   log file entries and storing of the data array to temporary file
        !
        ! *****************************************************************************
        !
        !   SUBROUTINES/FUNCTIONS CALLED
        !
        !   SwStartAndCount
        !   PrintToLog
        !   int2str
        !   SWStop
        !
        ! =============================================================================

        call PrintToLog("StoreField", 1)

        if (cSpace%cGrid%iDistr /= 0) then
            write (acTemp, '("Error occurred during the execution of StorePlane, aborting program")'); 
            call PrintToLog(acTemp, -1); 
            write (acTemp, '("the grid provided as input is not in distribution 0, but in distribution ", I3, "")') cSpace%cGrid%iDistr; 
            call PrintToLog(acTemp, -1); 
            stop
        end if

        call SwStartAndCount(cswDisk)

        !Check whether there are any positions in the FIRST z-plane on this processor
        !        !Check whether there are any positions in the last z-plane on this processor
        do iLi = 0, cSpace.cGrid.iD0LocN - 1
            !                if (cSpace.cGrid.aiD0Loc(1+iLi, 3) == cSpace.iDimZ-1) then
            if (cSpace.cGrid.aiD0Loc(1 + iLi, 3) == 0) then
                exit
            end if
        end do

        !If so, start from beginning position and save the FIRST z-plane
        !        !If so, start from beginning position and save the last z-plane
        if (iLi < cSpace.cGrid.iD0LocN) then

            do iLi2 = iLi, cSpace.cGrid.iD0LocN - 1
                if (cSpace.cGrid.aiD0Loc(1 + iLi2, 3) /= 0) then
                    exit
                end if
            end do

            iStart = iLi*cSpace%cGrid%iD0TL
            iLength = (iLi2 - iLi)*cSpace%cGrid%iD0TL
            !                iLength = (cSpace.cGrid.iD0LocN-iLi)*cSpace%cGrid%iD0TL
            !                        !iLength assumes that from iStart, the rest of the space belongs to the last slice...

            ! Write field to disk
            call PrintToLog('Write P_lin  to disk', 2); 
            call SwStartandCount(cswDiskAcces)
            acFileName = trim(sOutputDir)//acStoredFieldName//int2str(cSpace%cGrid%iProcID); 
            open (unit=iExportUNIT, file=trim(acFileName), status="REPLACE", form="UNFORMATTED", iostat=iErr)
            if (iErr /= 0) then
                write (acTemp, "('Error in StorePlane, file ',A20,'does not exist')") acStoredFieldName
                call PrintToLog(acTemp, -1)
                stop
            end if
            iLast = int(iLength/iBlockL); 
            do iLi = 0, iLast - 1
                write (unit=iExportUNIT, iostat=iErr) cSpace%cGrid%parD0(1 + iStart + iLi*iBlockL:iStart + (iLi + 1)*iBlockL); 
            end do
            !                write (unit=iExportUNIT, iostat=iErr) cSpace%cGrid%parD0(1+iStart+iLast*iBlockL:cSpace%cGrid%iD0LocSize);
            write (unit=iExportUNIT, iostat=iErr) cSpace%cGrid%parD0(1 + iStart + iLast*iBlockL:iStart + iLength); 
            close (unit=iExportUNIT); 
            call SwStop(cswDiskAcces)

        end if

        call SWStop(cswDisk)

    END SUBROUTINE StorePlane

    SUBROUTINE LoadPlane(cSpace, acStoredFieldName)

        ! =============================================================================
        !
        !   Programmer: Koos Huijssen
        !
        !   Language: Fortran 90
        !
        !   Version Date    Comment
        !   ------- -----   -------
        !   1.0     090505  Original code (KH)
        !
        ! *****************************************************************************
        !
        !   DESCRIPTION
        !
        !   The subroutine LoadPlane loads a plane in Z of the field stored in a temp
        !   file to the first plane in Z of the field in a given space. The data array
        !   is in Distr. 0, it is assumed that all other information in the space is
        !   correct. The storing was done in a blockwise fashion.
        !
        ! *****************************************************************************
        !
        !   INPUT/OUTPUT PARAMETERS
        !
        !   cSpace             io   type(space)  Space to which the data at the plane
        !                                        needs to be loaded.
        !   acStoredFieldName  io   char         Filename of the temp storage file.
        !
        type(Space), intent(inout)::            cSpace
        character(*), intent(in):: acStoredFieldName

        ! *****************************************************************************
        !
        !   LOCAL PARAMETERS
        !
        !   iErr            i8b    Error number
        !   iStart          i8b    Start index of the plane
        !   iLength         i8b    Length of the data array of the plane
        !   iLi             i8b    Loop counter
        !   iLi2            i8b    Loop counter
        !   iLast           i8b    Index of the last full block
        !   iBlockL         i8b    Size of the blocks
        !   parBuffer       dp     Temporary buffer for a block load
        !   acFilename      char   Total filename including path and suffix
        !   acTemp          char   Temporary char array for output log messages
        !
        integer(i8b)::                          iErr, iStart, iLength
        integer(i8b)::                          iLi, iLi2, iLast; 
        integer(i8b), parameter::               iBlockL = 2**36; 
        real(dp), allocatable::                                        parBuffer(:)
        character(len=1024)::                   acFileName
        character(len=1024)::                   acTemp; 
        ! *****************************************************************************
        !
        !   I/O
        !
        !   log file entries and loading of the data array from temporary file
        !
        ! *****************************************************************************
        !
        !   SUBROUTINES/FUNCTIONS CALLED
        !
        !   SwStartAndCount
        !   PrintToLog
        !   int2str
        !   SWStop
        !
        ! =============================================================================

        call PrintToLog("LoadPlane", 1)

        if (cSpace%cGrid%iDistr /= 0) then
            write (acTemp, '("Error occurred during the execution of LoadPlane, aborting program")'); 
            call PrintToLog(acTemp, -1); 
            write (acTemp, '("the grid provided as input is not in distribution 0, but in distribution ", I3, "")') cSpace%cGrid%iDistr; 
            call PrintToLog(acTemp, -1); 
            stop
        end if

        call SwStartAndCount(cswDisk)

        !Check whether there are any positions in the FIRST z-plane on this processor
        !        !Check whether there are any positions in the last z-plane on this processor
        do iLi = 0, cSpace.cGrid.iD0LocN - 1
            !                if (cSpace.cGrid.aiD0Loc(1+iLi, 3) == cSpace.iDimZ-1) then
            if (cSpace.cGrid.aiD0Loc(1 + iLi, 3) == 0) then
                exit
            end if
        end do

        !If so, start from beginning position and load the last z-plane
        if (iLi < cSpace.cGrid.iD0LocN) then

            do iLi2 = iLi, cSpace.cGrid.iD0LocN - 1
                if (cSpace.cGrid.aiD0Loc(1 + iLi2, 3) /= 0) then
                    exit
                end if
            end do

            iStart = iLi*cSpace%cGrid%iD0TL
            iLength = (iLi2 - iLi)*cSpace%cGrid%iD0TL
            !                iLength = (cSpace.cGrid.iD0LocN-iLi)*cSpace%cGrid%iD0TL
            !                        !iLength assumes that starting from iStart, the rest of the space belongs to the last slice...

            ! We read the original values and add them to cSpace
            allocate (parBuffer(iBlockL)); 
            call PrintToLog("Buffer allocated", 3)
            call PrintToLog('Read Stored field from disk', 2); 
            call SwStartAndCount(cswDiskAcces)
            acFileName = trim(sOutputDir)//acStoredFieldName//int2str(cSpace%cGrid%iProcID); 
            open (unit=iExportUNIT, file=trim(acFileName), status="OLD", form="UNFORMATTED", iostat=iErr)
            if (iErr /= 0) then
                write (acTemp, "('Error in LoadPlane, file ',A,' does not exist')") trim(acStoredFieldName)
                call PrintToLog(acTemp, -1)
                stop
            end if
            iLast = int(iLength/iBlockL); 
            do iLi = 0, iLast - 1
                read (unit=iExportUNIT, iostat=iErr) parBuffer; 
                cSpace%cGrid%parD0(1 + iStart + iLi*iBlockL:iStart + (iLi + 1)*iBlockL) = parBuffer
            end do
            !                read (unit=iExportUNIT, iostat=iErr) parBuffer(1:cSpace%cGrid%iD0LocSize-iLast*iBlockL);
            !                cSpace%cGrid%parD0(1+iStart+iLast*iBlockL:cSpace%cGrid%iD0LocSize) = parBuffer(1:cSpace%cGrid%iD0LocSize-iLast*iBlockL-iStart)
            read (unit=iExportUNIT, iostat=iErr) parBuffer(1:iLength - iLast*iBlockL); 
            cSpace%cGrid%parD0(1 + iStart + iLast*iBlockL:iStart + iLength) = parBuffer(1:iLength - iLast*iBlockL)

            close (unit=iExportUNIT); 
            call SwStop(cswDiskAcces)

            deallocate (parBuffer); 
        end if

        call SWStop(cswDisk)

        call PrintToLog("Free buffer", 3)

    END SUBROUTINE LoadPlane

    SUBROUTINE ZeroFirstPlane(cSpace)

        ! =============================================================================
        !
        !   Programmer: Koos Huijssen
        !
        !   Language: Fortran 90
        !
        !   Version Date    Comment
        !   ------- -----   -------
        !   1.0     090505  Original code (KH)
        !
        ! *****************************************************************************
        !
        !   DESCRIPTION
        !
        !   The subroutine ZeroFirstPlane puts the first plane in the Z-dimension in a
        !   Field space to zero. This is important if this field space is used as a
        !   contrast source to another space, and if at a different moment in the
        !   program the first plane of this space is be used as a pressure plane to
        !   account for the contrast sources at and before this plane. If we would not
        !   put it to zero in the contrast source to field computation, it would be
        !   accounted for twice.
        !
        ! *****************************************************************************
        !
        !   INPUT/OUTPUT PARAMETERS
        !
        !   cSpace   io   type(space)  Space in which the first plane in Z is put to
        !                              zero
        !
        type(Space), intent(inout)::            cSpace

        ! *****************************************************************************
        !
        !   LOCAL PARAMETERS
        !
        !   iErr            i8b    Error number
        !   iStart          i8b    Start index of the plane
        !   iEnd            i8b    End index of the plane
        !   iLi             i8b    Loop counter
        !   iLi2            i8b    Loop counter
        !   acTemp          char   Temporary char array for output log messages
        !
        integer(i8b)::                          iErr, iStart, iEnd
        integer(i8b)::                          iLi, iLi2
        character(len=1024)::                   acTemp

        ! *****************************************************************************
        !
        !   I/O
        !
        !   log file entries
        !
        ! *****************************************************************************
        !
        !   SUBROUTINES/FUNCTIONS CALLED
        !
        !   PrintToLog
        !
        ! =============================================================================

        call PrintToLog("ZeroFirstPlane", 1)

        if (cSpace%cGrid%iDistr /= 0) then
            write (acTemp, '("Error occurred during the execution of ZeroFirstPlane, aborting program")'); 
            call PrintToLog(acTemp, -1); 
            write (acTemp, '("the grid provided as input is not in distribution 0, but in distribution ", I3, "")') cSpace%cGrid%iDistr; 
            call PrintToLog(acTemp, -1); 
            stop
        end if

        !Check whether there are any positions in the first z-plane on this processor
        do iLi = 0, cSpace.cGrid.iD0LocN - 1
            if (cSpace.cGrid.aiD0Loc(1 + iLi, 3) == 0) then
                exit
            end if
        end do

        !If so, start from beginning position and load the last z-plane
        if (iLi < cSpace.cGrid.iD0LocN) then

            do iLi2 = iLi, cSpace.cGrid.iD0LocN - 1
                if (cSpace.cGrid.aiD0Loc(1 + iLi2, 3) /= 0) then
                    exit
                end if
            end do

            iStart = iLi*cSpace%cGrid%iD0TL
            iEnd = iLi2*cSpace%cGrid%iD0TL - 1

            do iLi = iStart, iEnd
                cSpace%cGrid%parD0(1 + iLi) = 0.0_dp
            end do

        end if

    END SUBROUTINE ZeroFirstPlane

    SUBROUTINE StoreSlices(pszFileName, pszVarName, cBeam, iBeam, bIsReal)

        ! =============================================================================
        !
        !   Programmer: Koos Huijssen
        !
        !   Language: Fortran 90
        !
        !   Version Date    Comment
        !   ------- -----   -------
        !   1.0     090505  Original code (KH)
        !
        ! *****************************************************************************
        !
        !   DESCRIPTION
        !
        !   The subroutine StoreSlices stores the slices from a space to output files.
        !   the slices to be stored are defined in the cModelParams structure. The
        !   filename is determined as
        !   'pszFilename//cModelParams%xyzslicedim//int2str(mmpos)//int2str(iBeam)//
        !   int2str(iProcID).h5'
        !
        ! *****************************************************************************
        !
        !   INPUT/OUTPUT PARAMETERS
        !
        !   pszFilename   i   char   Root of the filename to which the storage is done
        !   pszVarName    i   char   Variable name to be used in the file
        !   cBeam         i   type(space)  Space from which the data array needs to be
        !                                  stored.
        !   iBeam         i   i8b    Index number of the beam (closest to the source
        !                            is zero)
        !   bIsReal       i   lgt    Whether the information in a certain distribution
        !                            is real or not (this is not always clear,
        !                            especially in Distr. 1)
        !
        character(len=*), intent(in)::        pszFileName, pszVarName
        type(Space), INTENT(in) ::                cBeam
        integer(i8b), INTENT(in) ::                iBeam
        logical, intent(in) ::                        bIsReal

        ! *****************************************************************************
        !
        !   LOCAL PARAMETERS
        !
        !   i          i8b    loop counter over the number of slices to be saved
        !   mmpos      i8b    position in millimeters for the filename
        !   filename   char   Total filename including path and suffix
        !   acTemp     char   Temporary char array for output log messages
        !
        integer(i8b) :: i
        integer(i8b) :: mmpos
        character(LEN=1024) :: filename
        character(LEN=1024) :: acTemp

        ! *****************************************************************************
        !
        !   I/O
        !
        !   log file entries
        !
        ! *****************************************************************************
        !
        !   SUBROUTINES/FUNCTIONS CALLED
        !
        !   PrintToLog
        !   int2str
        !   ExportSlice
        !
        ! =============================================================================

        call PrintToLog("StoreSlices", 1)

        !        do i=1,cModelParams%numslices
        !                if ((cModelParams%xyzslicebeam(i)==iBeam).or.(cModelParams%xyzslicebeam(i)==-1)) then
        !                        mmpos =  int(cModelParams%xyzslicepos(i) * cMediumParams.c0/cModelParams.freq0 * 1.0D3)
        !                        filename = trim(pszFileName)//'_'//cModelParams%xyzslicedim(i)//&
        !                                                        int2str(mmpos)//int2str(iBeam)
        !                        if (cModelParams%xyzslicedim(i)=='t') then
        !                                call ExportSlice(trim(filename),pszVarName,cBeam, &
        !                                        (/ cModelParams%xyzsliceindex(i), 0_i8b, 0_i8b, 0_i8b /), &
        !                                        (/ 1_i8b, cBeam%iDimX, cBeam%iDimY, cBeam%iDimZ /), &
        !                                        bIsReal);
        !                        elseif (cModelParams%xyzslicedim(i)=='x') then
        !                                call ExportSlice(trim(filename),pszVarName,cBeam, &
        !                                        (/ 0_i8b, cModelParams%xyzsliceindex(i), 0_i8b, 0_i8b /), &
        !                                        (/ cBeam%iDimT, 1_i8b, cBeam%iDimY, cBeam%iDimZ /), &
        !                                        bIsReal);
        !                        elseif (cModelParams%xyzslicedim(i)=='y') then
        !                                call ExportSlice(trim(filename),pszVarName,cBeam, &
        !                                        (/ 0_i8b, 0_i8b, cModelParams%xyzsliceindex(i), 0_i8b /), &
        !                                        (/ cBeam%iDimT, cBeam%iDimX, 1_i8b, cBeam%iDimZ /), &
        !                                        bIsReal);
        !                        elseif (cModelParams%xyzslicedim(i)=='z') then
        !                                filename = trim(pszFileName)//'_'//cModelParams%xyzslicedim(i)//int2str(mmpos)
        !                                call ExportSlice(trim(filename),pszVarName,cBeam, &
        !                                        (/ 0_i8b, 0_i8b, 0_i8b, cModelParams%xyzsliceindex(i) /), &
        !                                        (/ cBeam%iDimT, cBeam%iDimX, cBeam%iDimY, 1_i8b /), &
        !                                        bIsReal);
        !                        end if
        !                end if
        !        end do

    !!L.D.
        do i = 1, cModelParams%numslices
            if ((cModelParams%xyzslicebeam(i) == iBeam) .or. (cModelParams%xyzslicebeam(i) == -1)) then
                mmpos = int((i))
                filename = trim(pszFileName)//'_'//cModelParams%xyzslicedim(i)// &
                           int2str(mmpos)//int2str(iBeam)

                if (cModelParams%xyzslicedim(i) == 't') then
                    call ExportSlice(trim(filename), pszVarName, cBeam, &
                                     (/cModelParams%xyzsliceindex(i), 0_i8b, 0_i8b, 0_i8b/), &
                                     (/1_i8b, cBeam%iDimX, cBeam%iDimY, cBeam%iDimZ/), &
                                     cModelParams.xyzsliceindex, cBeam%iDimT, bIsReal); 
                elseif (cModelParams%xyzslicedim(i) == 'x') then
                    call ExportSlice(trim(filename), pszVarName, cBeam, &
                                     (/0_i8b, cModelParams%xyzsliceindex(i), 0_i8b, 0_i8b/), &
                                     (/cBeam%iDimT, 1_i8b, cBeam%iDimY, cBeam%iDimZ/), &
                                     cModelParams.xyzsliceindex, cBeam%iDimX, bIsReal); 
                elseif (cModelParams%xyzslicedim(i) == 'y') then
                    call ExportSlice(trim(filename), pszVarName, cBeam, &
                                     (/0_i8b, 0_i8b, cModelParams%xyzsliceindex(i), 0_i8b/), &
                                     (/cBeam%iDimT, cBeam%iDimX, 1_i8b, cBeam%iDimZ/), &
                                     cModelParams.xyzsliceindex, cBeam%iDimY, bIsReal); 
                elseif (cModelParams%xyzslicedim(i) == 'z') then
!                filename = trim(pszFileName)//'_'//cModelParams%xyzslicedim(i)//int2str(mmpos)  ! Commented by A.M to be able to load files the same way
                    call ExportSlice(trim(filename), pszVarName, cBeam, &
                                     (/0_i8b, 0_i8b, 0_i8b, cModelParams%xyzsliceindex(i)/), &
                                     (/cBeam%iDimT, cBeam%iDimX, cBeam%iDimY, 1_i8b/), &
                                     cModelParams.xyzsliceindex, cBeam%iDimZ, bIsReal); 
                end if
            end if
        end do
    !! L.D.

    END SUBROUTINE StoreSlices

    SUBROUTINE UpdateRRMSerror(cRRMSNorm, cSpace, iBeam, iIteration)

        ! =============================================================================
        !
        !   Programmer: Koos Huijssen
        !
        !   Language: Fortran 90
        !
        !   Version Date    Comment
        !   ------- -----   -------
        !   1.0     090505  Original code (KH)
        !
        ! *****************************************************************************
        !
        !   DESCRIPTION
        !
        !   The subroutine UpdateRRMSerror updates the cRRMSNorm structure with the
        !   RRMS error of the current beam and iteration. This RRMS error is obtained
        !   as the Relative Root Mean Square error of the current iteration of the beam
        !   with respect to the previous iteration of the same beam. This is therefore
        !   more of a convergence check than a real error measure, and if it converges
        !   it is more the RRMS error of the PREVIOUS iteration than that of the
        !   current one. Moreover, it doesn't say that the scheme converges to the
        !   CORRECT solution. Only that it converges. If the RRMS error keeps on
        !   reducing, then we have convergence; if the RRMS error stays constant at
        !   a large value (say 1) or if it increases, then we have divergence.
        !   To reduce the storage between the iterationsw, only the RRMS error of the
        !   LAST plane of the field in the z-dimension is determined; it is assumed
        !   that it is a sufficient measure for the convergence of the rest.
        !   The subroutine also updates the arPreviousCorrection array to the last
        !   plane in z of the current iteration.
        !   In the subroutine, the data on all processors is collected by processor 0,
        !   which stores the total values for all beams and all iterations > 0.
        !
        ! *****************************************************************************
        !
        !   INPUT/OUTPUT PARAMETERS
        !
        !   cRRMSNorm   io   type(RRMSNormtype)   RRMSNorm structure to be updated
        !   cSpace      i    type(space)  Space for which the RRMSNorm is determined
        !   iBeam       i    i8b    Index number of the beam (closest to the source
        !                            is zero)
        !   iIteration  i    i8b    Iteration number (0 is primary field solution)
        !
        type(RRMSNormtype), INTENT(inout) :: cRRMSNorm
        type(Space), INTENT(in) :: cSpace
        integer(i8b), INTENT(in) :: iBeam, iIteration

        ! *****************************************************************************
        !
        !   LOCAL PARAMETERS
        !
        !   rRRMSNumDenom    dp   1D array of size 2 filled with 1) the sum of the
        !                         square errors of the positions on this processor, and
        !                         2) the sum of the square of the previous iteration of
        !                         the positions on this processor
        !   arRRMSNumDenom   dp   2D array used only by processor 0, which gathers all
        !                         numerator and denominator values of all processors
        !   prRRMSNumDenom   dp   pointer to an rRRMSNumDenom array
        !   parRRMSNumDenom  dp   pointer to an arRRMSNumDenom array
        !   iProcIndex       i4b  loop counter over the processor IDs
        !   iErr             i4b  error number to be used in MPI interfacing
        !   aiStatus         i4b  status of the MPI_Wait command
        !   aiReqSent        i4b  send request identification
        !   aiReqRecv        i4b  receive request identification
        !   acTemp     char   Temporary char array for output log messages
        !
        real(dp), target ::                rRRMSNumDenom(2)
        real(dp), target ::                arRRMSNumDenom(2, cSpace%cGrid%iProcN) !only used by processor 0
        real(dp), pointer ::        prRRMSNumDenom(:), parRRMSNumDenom(:, :)
        character(1024) ::                acTemp
        integer(i4b) ::        iProcIndex, iErr
        integer(i4b) ::        aiStatus(MPI_STATUS_SIZE)
        integer(i4b) ::        aiReqSent, aiReqRecv(cSpace%cGrid%iProcN - 1); 
        ! *****************************************************************************
        !
        !   I/O
        !
        !   log file entries
        !
        ! *****************************************************************************
        !
        !   SUBROUTINES/FUNCTIONS CALLED
        !
        !   PrintToLog
        !   MPI_ISend
        !   MPI_IRecv
        !   MPI_Wait
        !
        ! =============================================================================

        call PrintToLog("UpdateRRMSerror", 1)

        !check whether cBeam is in distribution 0
        if (cSpace%cGrid%iDistr /= 0) then
            write (acTemp, '("Error occurred during the execution of UpdateRRMSerror, aborting program")'); 
            call PrintToLog(acTemp, -1); 
            write (acTemp, '("the grid is not in distribution 0, but in distribution ", I3)') cSpace%cGrid%iDistr; 
            call PrintToLog(acTemp, -1); 
            stop
        end if

        !        calculate RRMS error for this processor
        !        calculate numerator and denominator separately
        rRRMSNumDenom = 0.0_dp
        if (cRRMSNorm%iStart .NE. -1) then
            rRRMSNumDenom(1) = sum((cSpace%cGrid%parD0(1 + cRRMSNorm%iStart:cSpace.cGrid.iD0LocN*cSpace%cGrid%iD0TL) &
                                    - cRRMSNorm%arPreviouscorrection)**2)
            rRRMSNumDenom(2) = sum(cSpace%cGrid%parD0(1 + cRRMSNorm%iStart:cSpace.cGrid.iD0LocN*cSpace%cGrid%iD0TL)**2)
        end if

        if (cSpace%cGrid%iProcID .NE. 0) then
            ! communicate RRMS error to processor 0;
            ! communicate 0 if there is nothing on this processor

            prRRMSNumDenom => rRRMSNumDenom

            write (acTemp, '(" Send RRMS error")'); call PrintToLog(acTemp, 4); 
            write (acTemp, '(" Num, Denom = ", E12.3, ", ",E12.3)') rRRMSNumDenom(1), rRRMSNumDenom(2); call PrintToLog(acTemp, 4); 
            call MPI_ISend(prRRMSNumDenom(1), &
                           2, &
                           MPI_DOUBLE_PRECISION, &
                           0, &
                           int(cSpace%cGrid%iProcID), &
                           MPI_COMM_WORLD, &
                           aiReqSent, &
                           iErr); 
        else
            !        if ProcID=0, collect intermediate answers from all processors and calculate final RRMS error, write to LOGfile

            !First numerator/denominator couple is from Processor 0 itself
            arRRMSNumDenom(:, 1) = rRRMSNumDenom
            parRRMSNumDenom => arRRMSNumDenom

            write (acTemp, '(" Receive RRMS errors")'); call PrintToLog(acTemp, 4); 
            do iProcIndex = 1, cSpace%cGrid%iProcN - 1
                call MPI_IRecv(parRRMSNumDenom(1, 1 + iProcIndex), &
                               2, &
                               MPI_DOUBLE_PRECISION, &
                               int(iProcIndex), &
                               int(iProcIndex), &
                               MPI_COMM_WORLD, &
                               aiReqRecv(iProcIndex), &
                               iErr); 
            end do

            write (acTemp, '(" Synchronize")'); call PrintToLog(acTemp, 4); 
            ! We wait untill we have obtained and sent all the blocks
            do iProcIndex = 1, cSpace%cGrid%iProcN - 1
                call MPI_Wait(aiReqRecv(iProcIndex), aiStatus, iErr); 
                write (acTemp, '("ProcID= ,",I3,", Num, Denom = ", E12.3,", ",E12.3)') iProcIndex, arRRMSNumDenom(1, 1 + iProcIndex), arRRMSNumDenom(2, 1 + iProcIndex); 
                call PrintToLog(acTemp, 5); 
            end do

            cRRMSNorm%arRRMSNorm(1 + iBeam, iIteration) = sqrt(sum(arRRMSNumDenom(1, :))/sum(arRRMSNumDenom(2, :)))

            write (acTemp, '("RRMSNorm: ", E14.5)') cRRMSNorm%arRRMSNorm(1 + iBeam, iIteration)
            call PrintToLog(acTemp, 1)
        end if

        !Update Previous Correction to the current correction
        if (cRRMSNorm%iStart .NE. -1) then
            cRRMSNorm%arPreviouscorrection = cSpace%cGrid%parD0(1 + cRRMSNorm%iStart:cSpace.cGrid.iD0LocN*cSpace%cGrid%iD0TL)
        end if

    END SUBROUTINE UpdateRRMSerror

    SUBROUTINE ExportRRMSerror(cRRMSNorm, iProcID)

        ! =============================================================================
        !
        !   Programmer: Koos Huijssen
        !
        !   Language: Fortran 90
        !
        !   Version Date    Comment
        !   ------- -----   -------
        !   1.0     090505  Original code (KH)
        !
        ! *****************************************************************************
        !
        !   DESCRIPTION
        !
        !   The subroutine ExportRRMSerror exports the RRMSnorm structure to a file
        !   with filename 'sOutputRoot // "RRMSNorm_000.h5"'. This only happens if
        !   the processor is 0, as this processor has collected all RRMSNorm data
        !   for all processors, beams and iterations.
        !
        ! *****************************************************************************
        !
        !   INPUT/OUTPUT PARAMETERS
        !
        !   cRRMSNorm   io   type(RRMSNormtype)   RRMSNorm structure to be exported
        !
        type(RRMSNormtype), INTENT(inout) :: cRRMSNorm

        ! *****************************************************************************
        !
        !   LOCAL PARAMETERS
        !
        !   iProcID    i8b    ID number of the current processor
        !
        integer(i8b) :: iProcID

        ! *****************************************************************************
        !
        !   I/O
        !
        !   log file entries
        !
        ! *****************************************************************************
        !
        !   SUBROUTINES/FUNCTIONS CALLED
        !
        !   ExportArray
        !
        ! =============================================================================

        !If ProcID=0, write complete cRRMSNorm to file
        if ((iProcID == 0) .AND. (allocated(cRRMSNorm%arRRMSNorm))) then

            call ExportArray(trim(sOutputDir)//trim(sOutputRoot)//"RRMSNorm", &
                             "RRMSNorm", cRRMSNorm%arRRMSNorm, iProcID)

        end if

    END SUBROUTINE ExportRRMSerror

    ! %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    ! %
    ! %        CopyDataD2
    ! %
    ! %        Strips the margins in the T dimension used for the calculation of d/dz
    ! %
    ! %        NB
    ! %        - cGridSource and cGridDest should be in distribution 0
    ! %
    ! %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    SUBROUTINE CopyDataD1(cGridSource, cGridDest, iTMargin)

        ! =============================================================================
        !
        !   Programmer: Koos Huijssen
        !
        !   Language: Fortran 90
        !
        !   Version Date    Comment
        !   ------- -----   -------
        !   1.0     090505  Original code (KH)
        !
        ! *****************************************************************************
        !
        !   DESCRIPTION
        !
        !   The subroutine CopyDataD1 copies the data from one grid to a grid with
        !   a smaller time dimension. This subroutine is used for stripping the margin
        !   in T that was necessary for the d/dz finite difference evaluation. It only
        !   works for grids in distribution 0. The margin which has to be removed at
        !   the beginning of each time trace needs to be given as input.
        !
        ! *****************************************************************************
        !
        !   INPUT/OUTPUT PARAMETERS
        !
        !   cGridSource   io   type(Grid)   Source grid
        !   cGridDest     io   type(Grid)   Destination grid
        !   iTMargin      i    i8b          Margin at the beginning of the T dimension
        !
        type(Grid), intent(inout)::                cGridSource, cGridDest; 
        integer(i8b), intent(in)::                iTMargin; 
        ! *****************************************************************************
        !
        !   LOCAL PARAMETERS
        !
        !   iT           i8b   loop counter over all time instants in an xyz-position
        !   iXYZ         i8b   loop counter over all xyz-positions in the beam
        !   iIndSource   i8b   index of the source data array
        !   iIndDest     i8b   index of the destination data array
        !
        integer(i8b)::                                iT, iXYZ, iIndSource, iIndDest

        ! *****************************************************************************
        !
        !   I/O
        !
        !   log file entries
        !
        ! *****************************************************************************
        !
        !   SUBROUTINES/FUNCTIONS CALLED
        !
        !   PrintToLog
        !
        ! =============================================================================

        call PrintToLog("CopyDataD1", 4)

        ! Copy the contents of the trace without the T margin
        do iXYZ = 0, cGridDest.iD0LocN - 1
            do iT = 0, cGridDest.iD0TL - 1
                iIndSource = 1 + iXYZ*cGridSource.iD0IS + (iT + iTMargin)*cGridSource.iD0TS; 
                iIndDest = 1 + iXYZ*cGridDest.iD0IS + iT*cGridDest.iD0TS; 
                cGridDest.parD0(iIndDest) = cGridSource.parD0(iIndSource); 
            end do
        end do

    END SUBROUTINE CopyDataD1

    SUBROUTINE CopyDataD0(cGridSource, cGridDest)

        ! =============================================================================
        !
        !   Programmer: Koos Huijssen
        !
        !   Language: Fortran 90
        !
        !   Version Date    Comment
        !   ------- -----   -------
        !   1.0     090505  Original code (KH)
        !
        ! *****************************************************************************
        !
        !   DESCRIPTION
        !
        !   The subroutine CopyDataD1 copies the data from one grid to a grid with
        !   a smaller time dimension. This subroutine is used for stripping the margin
        !   in T that was necessary for the d/dz finite difference evaluation. It only
        !   works for grids in distribution 0. The margin which has to be removed at
        !   the end of each time trace needs to be given as input.
        !
        ! *****************************************************************************
        !
        !   INPUT/OUTPUT PARAMETERS
        !
        !   cGridSource   io   type(Grid)   Source grid
        !   cGridDest     io   type(Grid)   Destination grid
        !   iTMargin      i    i8b          Margin at the beginning of the T dimension
        !
        type(Grid), intent(inout)::                cGridSource, cGridDest; 
        ! *****************************************************************************
        !
        !   LOCAL PARAMETERS
        !
        !   iT           i8b   loop counter over all time instants in an xyz-position
        !   iXYZ         i8b   loop counter over all xyz-positions in the beam
        !   iIndSource   i8b   index of the source data array
        !   iIndDest     i8b   index of the destination data array
        !
        integer(i8b)::                                iT, iXYZ, iIndSource, iIndDest

        ! *****************************************************************************
        !
        !   I/O
        !
        !   log file entries
        !
        ! *****************************************************************************
        !
        !   SUBROUTINES/FUNCTIONS CALLED
        !
        !   PrintToLog
        !
        ! =============================================================================

        call PrintToLog("CopyDataD1", 4)

        ! Copy the contents of the trace without the T margin
        do iXYZ = 0, cGridSource.iD0LocN - 1
            do iT = 0, cGridSource.iD0TL - 1
                iIndSource = 1 + iXYZ*cGridSource.iD0IS + iT*cGridSource.iD0TS; 
                iIndDest = 1 + iXYZ*cGridDest.iD0IS + iT*cGridDest.iD0TS; 
                cGridDest.parD0(iIndDest) = cGridSource.parD0(iIndSource); 
            end do
        end do
    END SUBROUTINE CopyDataD0

    SUBROUTINE CopyDataD2(cGridSource, cGridDest, iXYZMargin)

        ! =============================================================================
        !
        !   Programmer: Koos Huijssen
        !
        !   Language: Fortran 90
        !
        !   Version Date    Comment
        !   ------- -----   -------
        !   1.0     090505  Original code (KH)
        !
        ! *****************************************************************************
        !
        !   DESCRIPTION
        !
        !   The subroutine CopyDataD2 copies the data from one grid to a grid with
        !   smaller dimensions in X,Y and Z. This subroutine is used for stripping the
        !   margins in X, Y and Z that were necessary for the d/dz finite difference
        !   evaluation. It only works for grids in distribution 2. The margin which has
        !   to be removed at the beginning the XYZ dimensions need to be given as
        !   input.
        !
        ! *****************************************************************************
        !
        !   INPUT/OUTPUT PARAMETERS
        !
        !   cGridSource   io   type(Grid)   Source grid
        !   cGridDest     io   type(Grid)   Destination grid
        !   iXYZMargin    i    i8b          Margin at the beginning of the XYZ
        !                                   dimensions
        !
        type(Grid), intent(inout)::                cGridSource, cGridDest; 
        integer(i8b), intent(in)::                iXYZMargin(:); 
        ! *****************************************************************************
        !
        !   LOCAL PARAMETERS
        !
        !   iT           i8b   loop counter over the t-axis in the beam
        !   iX           i8b   loop counter over the x-axis in the beam
        !   iY           i8b   loop counter over the y-axis in the beam
        !   iZ           i8b   loop counter over the z-axis in the beam
        !   iIndSource   i8b   index of the source data array
        !   iIndDest     i8b   index of the destination data array
        !
        integer(i8b)::                                iT, iX, iY, iZ, iIndSource, iIndDest

        ! *****************************************************************************
        !
        !   I/O
        !
        !   log file entries
        !
        ! *****************************************************************************
        !
        !   SUBROUTINES/FUNCTIONS CALLED
        !
        !   PrintToLog
        !
        ! =============================================================================

        call PrintToLog("CopyDataD2", 4)

        ! Copy the contents of the slices without the XYZ margin
        do iT = 0, cGridDest.iD2LocN - 1
            do iY = 0, cGridDest.iD2YL - 1
                do iX = 0, cGridDest.iD2XL - 1
                    iIndSource = 1 + iT*cGridSource.iD2IS + (iX + iXYZMargin(1))*cGridSource.iD2XS + (iY + iXYZMargin(2))*cGridSource.iD2YS + iXYZMargin(3)*cGridSource.iD2ZS; 
                    iIndDest = 1 + iT*cGridDest.iD2IS + iX*cGridDest.iD2XS + iY*cGridDest.iD2YS; 
                    do iZ = 0, cGridDest.iD2ZL - 1

                        cGridDest.pacD2(iIndDest) = cGridSource.pacD2(iIndSource); 
                        iIndSource = iIndSource + cGridSource.iD2ZS; 
                        iIndDest = iIndDest + cGridDest.iD2ZS; 
                    end do
                end do
            end do
        end do

    END SUBROUTINE CopyDataD2

    SUBROUTINE CopyToLargeDataD2(cGridSource, cGridDest, iXYZMargin) 

        ! =============================================================================
        !
        !   Programmer: Koos Huijssen
        !
        !   Language: Fortran 90
        !
        !   Version Date    Comment
        !   ------- -----   -------
        !   1.0     090505  Original code (KH)
        !
        ! *****************************************************************************
        !
        !   DESCRIPTION
        !
        !   The subroutine CopyDataD2 copies the data from one grid to a grid with
        !   smaller dimensions in X,Y and Z. This subroutine is used for stripping the
        !   margins in X, Y and Z that were necessary for the d/dz finite difference
        !   evaluation. It only works for grids in distribution 2. The margin which has
        !   to be removed at the beginning the XYZ dimensions need to be given as
        !   input.
        !
        ! *****************************************************************************
        !
        !   INPUT/OUTPUT PARAMETERS
        !
        !   cGridSource   io   type(Grid)   Source grid
        !   cGridDest     io   type(Grid)   Destination grid
        !   iXYZMargin    i    i8b          Margin at the beginning of the XYZ
        !                                   dimensions
        !
        type(Grid), intent(inout)::                cGridSource, cGridDest; 
        integer(i8b), intent(in)::                iXYZMargin(:); 
        ! *****************************************************************************
        !
        !   LOCAL PARAMETERS
        !
        !   iT           i8b   loop counter over the t-axis in the beam
        !   iX           i8b   loop counter over the x-axis in the beam
        !   iY           i8b   loop counter over the y-axis in the beam
        !   iZ           i8b   loop counter over the z-axis in the beam
        !   iIndSource   i8b   index of the source data array
        !   iIndDest     i8b   index of the destination data array
        !
        integer(i8b)::                                iT, iX, iY, iZ, iIndSource, iIndDest

        ! *****************************************************************************
        !
        !   I/O
        !
        !   log file entries
        !
        ! *****************************************************************************
        !
        !   SUBROUTINES/FUNCTIONS CALLED
        !
        !   PrintToLog
        !
        ! =============================================================================

        call PrintToLog("CopyDataD2", 4)

        ! Copy the contents of the slices without the XYZ margin
        do iT = 0, cGridSource.iD2LocN - 1
            do iY = 0, cGridSource.iD2YL - 1
                do iX = 0, cGridSource.iD2XL - 1
                    iIndDest= 1 + iT*cGridDest.iD2IS + (iX + iXYZMargin(1))*cGridDest.iD2XS + (iY + iXYZMargin(2))*cGridDest.iD2YS + iXYZMargin(3)*cGridDest.iD2ZS; 
                    iIndSource = 1 + iT*cGridSource.iD2IS + iX*cGridSource.iD2XS + iY*cGridSource.iD2YS; 
                    do iZ = 0, cGridSource.iD2ZL - 1

                        cGridDest.pacD2(iIndDest) = cGridSource.pacD2(iIndSource); 
                        iIndSource = iIndSource + cGridSource.iD2ZS; 
                        iIndDest = iIndDest + cGridDest.iD2ZS; 
                    end do
                end do
            end do
        end do

    END SUBROUTINE CopyToLargeDataD2 

    SUBROUTINE test_isnan(cSpace)

        ! =============================================================================
        !
        !   Programmer: Koos Huijssen
        !
        !   Language: Fortran 90
        !
        !   Version Date    Comment
        !   ------- -----   -------
        !   1.0     090505  Original code (KH)
        !
        ! *****************************************************************************
        !
        !   DESCRIPTION
        !
        !   The subroutine test_isnan tests whether there is a NaN or a very large
        !   number occurring in a space under test. This would be an indication of an
        !   error during calculation, and it would be a waste of resources to continue
        !   the evaluation.
        !
        ! *****************************************************************************
        !
        !   INPUT/OUTPUT PARAMETERS
        !
        !   cSpace   i   type(Space)   Space which needs to be tested
        !
        type(Space), INTENT(in) :: cSpace

        ! *****************************************************************************
        !
        !   LOCAL PARAMETERS
        !
        !   iLi   i8b   loop counter over all array elements
        !
        integer(i8b) :: iLi
        character(LEN=1024)::   acTemp; 
        ! *****************************************************************************
        !
        !   I/O
        !
        !   log file entries
        !
        ! *****************************************************************************
        !
        !   SUBROUTINES/FUNCTIONS CALLED
        !
        !   PrintToLog
        !
        ! =============================================================================

        if (cSpace%cGrid%iDistr == 0) then
            if (.not. ALL(cSpace%cGrid%ParD0 < 1E100_dp)) then
                !write (acTemp, '("Error: NaN or very large number encountered!!!" ,I3," - Processor checking in.")') cSpace.cGrid.iProcID; call PrintToLog(acTemp, 0);
                call PrintToLog("Error: NaN or very large number encountered!!!", 1)
                stop
            end if
        elseif (cSpace%cGrid%iDistr == 1) then

            if (.not. ALL(real(cSpace%cGrid%PacD1, dp) < 1E100_dp) .OR. .NOT. ALL(dimag(cSpace%cGrid%PacD1) < 1E100_dp)) then
                call PrintToLog("Error: NaN or very large number encountered!!!", 1)
                stop
            end if
        elseif (cSpace%cGrid%iDistr == 2) then
            if (.not. ALL(real(cSpace%cGrid%PacD2, dp) < 1E100_dp) .OR. .NOT. ALL(dimag(cSpace%cGrid%PacD2) < 1E100_dp)) then
                call PrintToLog("Error: NaN or very large number encountered!!!", 1)
                stop
            end if
        end if

    END SUBROUTINE test_isnan

    SUBROUTINE TwoRealValueCalculation(a, B, C)
        ! =============================================================================
        !
        !   Programmer: Libertario Demi
        !
        !   Language: Fortran 90
        !
        !   Version Date    Comment
        !   ------- -----   -------
        !   1.0     210512  Original code (LD)
        !
        ! *****************************************************************************
        !
        ! *****************************************************************************
        !
        !   INPUT/OUTPUT PARAMETERS
        !
        !   cSpace             io   type(space)  Space to which the stored data array
        !                                        is subtracted.
        !   acStoredFieldName  io   char         Filename of the temp storage file.
        !
        type(Space), intent(in)   ::            B, C
        integer(i8b)::                          iLi, Vectordimensions
        character(len=1024)::                   acTemp; 
        real(8), intent(inout):: a

        ! *****************************************************************************
        !
        !   I/O
        !
        !   log file entries and loading of the data array from temporary file
        !
        ! *****************************************************************************
        !
        !   SUBROUTINES/FUNCTIONS CALLED
        !
        !   SwStartAndCount
        !   PrintToLog
        !   int2str
        !   SWStop
        !
        ! =============================================================================

        a = 0

        call PrintToLog("TwoRealCalculation", 1)

        if (B%cGrid%iDistr /= 0) then
            write (acTemp, '("Error occurred during the execution of TwoRealCalculation, aborting program")'); 
            call PrintToLog(acTemp, -1); 
            write (acTemp, '("the grid provided as input is not in distribution 0, but in distribution ", I3, "")') B%cGrid%iDistr; 
            call PrintToLog(acTemp, -1); 
            stop
        end if

        if (C%cGrid%iDistr /= 0) then
            write (acTemp, '("Error occurred during the execution of TwoRealCalculation, aborting program")'); 
            call PrintToLog(acTemp, -1); 
            write (acTemp, '("the grid provided as input is not in distribution 0, but in distribution ", I3, "")') B%cGrid%iDistr; 
            call PrintToLog(acTemp, -1); 
            stop
        end if

        Vectordimensions = size(B%cGrid%parD0)

        do iLi = 0, Vectordimensions - 1

            a = a + (B%cGrid%parD0(1 + iLi)*C%cGrid%parD0(1 + iLi))
        end do

        call PrintToLog("TwoRealCalculation Finished", 3)

        a = 2*REAL(a); 
    END SUBROUTINE TwoRealValueCalculation

    SUBROUTINE AbsoluteValueCalculation(a, B)
        ! =============================================================================
        !
        !   Programmer: Libertario Demi
        !
        !   Language: Fortran 90
        !
        !   Version Date    Comment
        !   ------- ------  -------
        !   1.0     210512  Original code (LD)
        !
        ! *****************************************************************************
        !
        !   DESCRIPTION
        !
        !   The subroutine AbsoluteValueCalculation calculates the summation of the
        !   auto-inner product
        !
        ! *****************************************************************************
        !
        !   INPUT/OUTPUT PARAMETERS
        !
        !   cSpace             io   type(space)  Space to which the stored data array
        !                                        is subtracted.
        !

        type(Space), intent(in)   ::            B
        integer(i8b)::                          iLi, Vectordimensions
        character(len=1024)::                   acTemp; 
        real(8), intent(inout):: a

        ! *****************************************************************************
        !
        ! =============================================================================
        a = 0

        call PrintToLog("AbsoluteValueCalculation", 1)

        if (B%cGrid%iDistr /= 0) then
            write (acTemp, '("Error occurred during the execution of AbsoluteValueCalculation, aborting program")'); 
            call PrintToLog(acTemp, -1); 
            write (acTemp, '("the grid provided as input is not in distribution 0, but in distribution ", I3, "")') B%cGrid%iDistr; 
            call PrintToLog(acTemp, -1); 
            stop
        end if

        Vectordimensions = size(B%cGrid%parD0)

        do iLi = 0, Vectordimensions - 1

            a = a + (B%cGrid%parD0(1 + iLi)*B%cGrid%parD0(1 + iLi))
        end do

        call PrintToLog("AbsoluteValueCalculation Finished", 3)

    END SUBROUTINE AbsoluteValueCalculation

END MODULE ParnacIterStepMain
